#region << Using Directives >>
using System;
using Volpe.Cafe.Data;
using Volpe.Cafe.EIS.Data;
using Volpe.Cafe.EIS.IO.Reporting;
using Volpe.Cafe.EIS.Settings;
using Volpe.Cafe.IO.Reporting.XL;
using Volpe.Cafe.Model;
using Volpe.Cafe.Settings;
using Volpe.Cafe.Utils;
#endregion
namespace Volpe.Cafe.EIS.Model
{
    [Serializable]
    [ModelDescription("Compliance Model with EIS", "Runs a regular compliance model, including additional analysis necessary " +
        "for the Environmental Impact Statement.", 1.0F)]
    public sealed class EISModel : Compliance
    {
        #region 
        static EISModel()
        {
            FuelTypes = new FuelType[] {FuelType.Gasoline, FuelType.Diesel, FuelType.Ethanol85, FuelType.Electricity, FuelType.Hydrogen};
            FuelTypeCount = FuelTypes.Length;
        }
        public EISModel() : base() { }
        #endregion
        #region 
        #region 
        public override ICompliance CreateNew()
        {
            return new EISModel();
        }
        protected override XlReportGenerator CreateXlReportGenerator()
        {
            return new XlEISReportGenerator(this);
        }
        protected override string Validate()
        {
            string message = base.Validate();
            if (message != string.Empty) { return message; }
            EISSettings eisSettings = (EISSettings)this.Settings.GetRuntimeSettings(typeof(EISSettings));
            int missing = 0;
            if (eisSettings.EISParameters        == null) { missing |= 1; }
            if (eisSettings.EISTailpipeEmissions == null) { missing |= 2; }
            if (missing == 0) { return string.Empty; }
            else
            {
                if      (missing == 1) { message = "EIS Parameters"; }
                else if (missing == 2) { message = "EIS Tailpipe Emission Rates"; }
                else if (missing == 3) { message = "EIS Parameters and EIS Tailpipe Emission Rates"; }
                message = "The following errors have occured during initialization:\n\n    The " + message +
                    " are missing from the modeling settings.\n\nThe model will not run.";
                return message;
            }
        }
        #endregion
        #region 
        protected override bool PreProcess()
        {
            if (!base.PreProcess()) { return false; }
            this._eisRunning = false;
            this._eisSettings = (EISSettings)this.Settings.GetRuntimeSettings(typeof(EISSettings));
            this._parameters           = this.    Settings.Parameters;
            this._eisParameters        = this._eisSettings.EISParameters;
            this._eisTailpipeEmissions = this._eisSettings.EISTailpipeEmissions;
            this._eisMinYear = Parameters.FuelEconomyDataValues.MinYear;
            this._eisMaxYear = Parameters.FuelEconomyDataValues.MaxYear;
            this._eisData = new EISModelData(this.Settings.Scenarios.Count, this._eisMinYear, this._eisMaxYear);
            return true;
        }
        #endregion
        protected override void PostProcessScenario(Scenario scen)
        {
            if (this.AbortRequested) { return; }
            this._eisRunning = true;
            this.CalculateScenarioData(scen, this._eisData.Cars[scen.Index], true);
            this.CalculateScenarioData(scen, this._eisData.Trks[scen.Index], false);
            if (scen.IsBaseline)
            {   
                this.CalculateScenarioData(null, this._eisData.CarsNoCAFE, true);
                this.CalculateScenarioData(null, this._eisData.TrksNoCAFE, false);
            }
            this._eisRunning = false;
            base.PostProcessScenario(scen);
        }
        void CalculateScenarioData(Scenario scen, EISAnalysisData eisData, bool isCar)
        {
            EISParameters.NoCAFEData            noCAFE     = this._eisSettings.EISParameters.NoCAFE;
            Parameters   .FuelEconomyDataValues wCAFE      = this._parameters .FuelEconomyData;
            bool                                overcomply = this.Settings.OperatingModes.Overcomply;
            this.UpdateFuelEconomyLevels(scen, eisData, isCar, wCAFE, noCAFE, overcomply, this._eisSettings.EISParameters.Overcompliance);
            for (int i = this._eisMinYear; i <= this._eisMaxYear; i++)
            {   
                this.InitializeFuelPriceArrays(i);
                int       yrIndex       = i - this._eisMinYear;
                FuelValue fuelShare     = eisData.FuelShare[yrIndex];
                FuelValue fuelEconomy   = eisData.OnRoadFuelEconomy[yrIndex];
                FuelValue energyDensity = this._parameters.FuelProperties.EnergyDensity;
                Estimates fpEstimates   = this.Settings.OperatingModes.FuelPriceEstimates;
                for (int j = 0; j < Interaction.CY; j++)
                {
                    double fleet = this.GetFleet      (i, j, isCar);
                    double miles = this.GetMilesDriven(   j, isCar);
                    double vmtGrowth = this._parameters.EconomicData.GetVMTGrowthFactor(fpEstimates, i, j);
                    double cpmRate = this.GetCPMRate(scen, i, j, eisData, isCar);
                    eisData.Fleet[yrIndex, yrIndex + j] = fleet;
                    for (int f = 0; f < FuelTypeCount; f++)
                    {
                        FuelType fuel = FuelTypes[f];
                        if (fuelShare[fuel] == 0) { continue; }
                        double vmt = fuelShare[fuel] * fleet * miles * vmtGrowth / 1e9 * (1 + cpmRate);
                        eisData.VMT[yrIndex, yrIndex + j][fuel] = vmt;
                        double gal = vmt / fuelEconomy[fuel];
                        eisData.Gallons[yrIndex, yrIndex + j][fuel] = gal;
                        double btu = 0, energyUse = 0;
                        if (fuel == FuelType.Hydrogen || fuel == FuelType.Electricity)
                        {
                            btu       = (gal * energyDensity.Gasoline);
                            energyUse = (btu / energyDensity[fuel]);
                        }
                        else
                        {
                            btu       = (gal * energyDensity[fuel]);
                            energyUse = (gal);
                        }
                        Mobile6Class m6 = (isCar) ? Mobile6Class.LDV : Mobile6Class.LDT1;
                        this.CalculateEmissions(eisData.EmissionsCO          , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.CO          , this._eisTailpipeEmissions.CO          [i, fuel, m6], false, false);
                        this.CalculateEmissions(eisData.EmissionsVOC         , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.VOC         , this._eisTailpipeEmissions.VOC         [i, fuel, m6], false, false);
                        this.CalculateEmissions(eisData.EmissionsNOX         , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.NOX         , this._eisTailpipeEmissions.NOX         [i, fuel, m6], false, false);
                        this.CalculateEmissions(eisData.EmissionsPM          , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.PM          , this._eisTailpipeEmissions.PM          [i, fuel, m6], false, false);
                        this.CalculateEmissions(eisData.EmissionsSOX         , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.SOX         , null                                                , true , false);
                        this.CalculateEmissions(eisData.EmissionsCO2         , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.CO2         , null                                                , false, true );
                        this.CalculateEmissions(eisData.EmissionsAcetaldehyde, yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.Acetaldehyde, this._eisTailpipeEmissions.Acetaldehyde[i, fuel, m6], false, false);
                        this.CalculateEmissions(eisData.EmissionsAcrolein    , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.Acrolein    , this._eisTailpipeEmissions.Acrolein    [i, fuel, m6], false, false);
                        this.CalculateEmissions(eisData.EmissionsBenzene     , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.Benzene     , this._eisTailpipeEmissions.Benzene     [i, fuel, m6], false, false);
                        this.CalculateEmissions(eisData.EmissionsButadiene   , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.Butadiene   , this._eisTailpipeEmissions.Butadiene   [i, fuel, m6], false, false);
                        this.CalculateEmissions(eisData.EmissionsFormaldehyde, yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.Formaldehyde, this._eisTailpipeEmissions.Formaldehyde[i, fuel, m6], false, false);
                        this.CalculateEmissions(eisData.EmissionsCH4         , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.CH4         , this._eisTailpipeEmissions.CH4         [i, fuel, m6], false, false);
                        this.CalculateEmissions(eisData.EmissionsDPM         , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.DieselPM    , this._eisTailpipeEmissions.DieselPM    [i, fuel, m6], false, false);
                        this.CalculateEmissions(eisData.EmissionsMTBE        , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.MTBE        , this._eisTailpipeEmissions.MTBE        [i, fuel, m6], false, false);
                        this.CalculateEmissions(eisData.EmissionsN2O         , yrIndex, j, fuel, btu, energyUse, vmt, this._parameters.UpstreamEmissions.N2O         , this._eisTailpipeEmissions.N2O         [i, fuel, m6], false, false);
                    } 
                } 
            } 
        }
        void CalculateEmissions(EISEmissionsData em, int yrIndex, int age, FuelType fuel,
            double btu, double energyUse, double vmt, FuelValue ue, double[] te, bool isSOX, bool isCO2)
        {
            const double kgToTons = 1102.31131; 
            const double kgToMMT  = 1e-3;       
            Parameters.FuelPropertiesValues fuelProp = this._parameters.FuelProperties;
            double upstreamEm, tailpipeEm;
            if (isCO2)
            {   
                upstreamEm = ue[fuel] * btu / 1e6 * kgToMMT;
                tailpipeEm = fuelProp.MassDensity[fuel] * fuelProp.CarbonContent[fuel] * 44 / 12 * energyUse * kgToMMT;
            }
            else if (isSOX)
            {   
                upstreamEm = ue[fuel] * btu / 1e6 * kgToTons;
                tailpipeEm = fuelProp.SOXRate[fuel] * energyUse * kgToTons;
            }
            else
            {   
                upstreamEm = ue[fuel] * btu / 1e6 * kgToTons;
                tailpipeEm = (te == null) ? 0 : te[Math.Min(age, te.Length - 1)] * vmt * kgToTons;
            }
            em.Upstream[yrIndex, yrIndex + age] += upstreamEm;
            em.Tailpipe[yrIndex, yrIndex + age] += tailpipeEm;
            em.Total   [yrIndex, yrIndex + age] += (upstreamEm + tailpipeEm);
        }
        double GetCPMRate(Scenario scen, int year, int age, EISAnalysisData eisData, bool isCar)
        {
            int yrIndex = year - this._eisMinYear;
            int hIndex = this._parameters.EconomicData.VMTBaseYear - Parameters.FuelEconomyDataValues.MinYear - age;
            if (hIndex < 0) { hIndex = 0; }
            FuelValue hFE = (isCar) ? this._parameters.FuelEconomyData.CarsFuelEconomy[hIndex] : this._parameters.FuelEconomyData.TrksFuelEconomy[hIndex];
            FuelValue hFS = (isCar) ? this._parameters.FuelEconomyData.CarsFuelShare  [hIndex] : this._parameters.FuelEconomyData.TrksFuelShare  [hIndex];
            FuelValue sFE = eisData.OnRoadFuelEconomy[yrIndex];
            FuelValue sFS = eisData.FuelShare        [yrIndex];
            FuelValue gap = this._parameters.EconomicData.TestVsOnRoadGap;
            double hCPM = 0, sCPM = 0;
            for (int f = 0; f < FuelTypeCount; f++)
            {
                FuelType fuel = FuelTypes[f];
                if (hFS[fuel] > 0) { hCPM += (this.RetailFPBaseVMT[f] / (hFE[fuel] * (1 - gap[fuel])) * hFS[fuel]); }
                if (sFS[fuel] > 0) { sCPM += (this.RetailFP[f][age] / sFE[fuel] * sFS[fuel]); }
            }
            return (sCPM / hCPM - 1) * this._parameters.EconomicData.ReboundEffect;
        }
        double GetFleet(int eisYear, int age, bool isCar)
        {
            if (eisYear <= EISParameters.FleetData.MaxYear)
            {   
                return (isCar) ?
                    this._eisParameters.Fleet.CarFleet[eisYear - EISParameters.FleetData.MinYear][age] :
                    this._eisParameters.Fleet.TrkFleet[eisYear - EISParameters.FleetData.MinYear][age];
            }
            else
            {   
                double sales = (isCar) ?
                    this._eisParameters.Sales.CarSales[eisYear - EISParameters.SalesData.MinYear] :
                    this._eisParameters.Sales.TrkSales[eisYear - EISParameters.SalesData.MinYear];
                double survRate = (isCar) ?
                    this._parameters.VehicleAgeData.SurvivalCars  [age] :
                    this._parameters.VehicleAgeData.SurvivalAllLTs[age];
                return sales * survRate;
            }
        }
        double GetMilesDriven(int age, bool isCar)
        {
            return (isCar) ? this._parameters.VehicleAgeData.MilesCars[age] : this._parameters.VehicleAgeData.MilesAllLTs[age];
        }
        void UpdateFuelEconomyLevels(Scenario scen, EISAnalysisData eisData, bool isCar, Parameters.FuelEconomyDataValues wCAFE,
            EISParameters.NoCAFEData noCAFE, bool overcomply, EISParameters.OvercomplianceData feOvercomply)
        {
            if (scen == null)
            {   
                Array.Copy((isCar) ? noCAFE.CarFuelEconomy : noCAFE.TrkFuelEconomy, eisData.RatedFuelEconomy, eisData.NumYears);
                Array.Copy((isCar) ? noCAFE.CarFuelShare   : noCAFE.TrkFuelShare  , eisData.FuelShare       , eisData.NumYears);
            }
            else
            {   
                Array.Copy((isCar) ? wCAFE.CarsFuelEconomy : wCAFE.TrksFuelEconomy, eisData.RatedFuelEconomy, eisData.NumYears);
                Array.Copy((isCar) ? wCAFE.CarsFuelShare   : wCAFE.TrksFuelShare  , eisData.FuelShare       , eisData.NumYears);
            }
            FuelValue gap = this.Settings.Parameters.EconomicData.TestVsOnRoadGap;
            for (int i = this._eisMinYear; i <= this._eisMaxYear; i++)
            {
                int yrIndex = i - this._eisMinYear;
                for (int k = 0; k < FuelTypeCount; k++)
                {
                    FuelType fuel = FuelTypes[k];
                    if (eisData.RatedFuelEconomy[yrIndex][fuel] == 0) { continue; }
                    eisData.OnRoadFuelEconomy[yrIndex][fuel] = eisData.RatedFuelEconomy[yrIndex][fuel] * (1 - gap[fuel]);
                }
            }
            if (scen == null) { return; }
            Industry[] scenData = this.GetData(scen);
            for (int i = this.MinYear; i <= this.MaxYear; i++)
            {   
                ModelYear year = this.ModelYears[i - this.MinYear];
                Industry data = scenData[year.Index];
                FuelValue ratedFE, onRoadFE, fs;
                this.ComputeIndustryFuelValues(data, year, isCar, out ratedFE, out onRoadFE, out fs);
                int index = i - this._eisMinYear;
                eisData.RatedFuelEconomy [index] = ratedFE;
                eisData.OnRoadFuelEconomy[index] = onRoadFE;
                eisData.FuelShare        [index] = fs;
                if (i == this.MaxYear)
                {
                    FuelValue[] baseFE = (isCar) ? this._eisData.Cars[0].RatedFuelEconomy : this._eisData.Trks[0].RatedFuelEconomy;
                    FuelValue[] baseGr = (isCar) ? feOvercomply.CarBaseGrowth : feOvercomply.TrkBaseGrowth;
                    FuelValue[] scenGr = (isCar) ? feOvercomply.CarScenGrowth : feOvercomply.TrkScenGrowth;
                    i++;
                    for (; i <= this._eisMaxYear; i++)
                    {
                        index = i - this._eisMinYear;
                        if (!overcomply)
                        {   
                            eisData.RatedFuelEconomy[index] = eisData.RatedFuelEconomy[index - 1];
                        }
                        else if (scen.IsBaseline)
                        {   
                            eisData.RatedFuelEconomy[index] = eisData.RatedFuelEconomy[index - 1] * (FuelValue.One + baseGr[index]);
                        }
                        else
                        {   
                            eisData.RatedFuelEconomy[index] = eisData.RatedFuelEconomy[index - 1] * (FuelValue.One + scenGr[index]);
                            FuelValue scenFE = new FuelValue();
                            scenFE[FuelType.Gasoline   ] = Math.Max(eisData.RatedFuelEconomy[index].Gasoline   , baseFE[index].Gasoline   );
                            scenFE[FuelType.Diesel     ] = Math.Max(eisData.RatedFuelEconomy[index].Diesel     , baseFE[index].Diesel     );
                            scenFE[FuelType.Ethanol85  ] = Math.Max(eisData.RatedFuelEconomy[index].Ethanol85  , baseFE[index].Ethanol85  );
                            scenFE[FuelType.Electricity] = Math.Max(eisData.RatedFuelEconomy[index].Electricity, baseFE[index].Electricity);
                            scenFE[FuelType.Hydrogen   ] = Math.Max(eisData.RatedFuelEconomy[index].Hydrogen   , baseFE[index].Hydrogen   );
                            eisData.RatedFuelEconomy[index] = scenFE;
                        }
                        eisData.OnRoadFuelEconomy[index] = eisData.RatedFuelEconomy[index] * (FuelValue.One - gap);
                        eisData.FuelShare        [index] = fs;
                    }
                }
            } 
        }
        void ComputeIndustryFuelValues(Industry data, ModelYear year, bool isCar, out FuelValue ratedFE, out FuelValue onRoadFE, out FuelValue fuelShares)
        {
            FuelValue salesSum = FuelValue.Zero;
            FuelValue feSum    = FuelValue.Zero;
            for (int i = 0, mfrCount = data.ManufacturerCount; i < mfrCount; i++)
            {
                Manufacturer mfr = data.Manufacturers[i];
                for (int j = 0, vehCount = mfr.VehicleCount; j < vehCount; j++)
                {
                    Vehicle veh = mfr.Vehicles[j];
                    if ((isCar && veh.Mobile6Class == Mobile6Class.LDV) || (!isCar && veh.Mobile6Class != Mobile6Class.LDV))
                    {
                        double vehSales = veh.Description.Sales[year.Index];
                        FuelValue vehFE = Effects.GetVehcleFE(veh, this._scen, year, this.Settings.Parameters);  
                        FuelValue vehFS = veh.Description.FuelShare;
                        for (int k = 0; k < FuelTypeCount; k++)
                        {
                            FuelType fuel = FuelTypes[k];
                            if (vehFS[fuel] == 0) { continue; }
                            salesSum[fuel] += vehSales * vehFS[fuel];
                            feSum   [fuel] += vehSales * vehFS[fuel] / vehFE[fuel];
                        }
                    }
                } 
            } 
            FuelValue gap = this.Settings.Parameters.EconomicData.TestVsOnRoadGap;
            ratedFE    = new FuelValue();
            onRoadFE   = new FuelValue();
            fuelShares = new FuelValue();
            for (int i = 0; i < FuelTypeCount; i++)
            {
                FuelType fuel = FuelTypes[i];
                if (salesSum[fuel] == 0) { continue; }
                onRoadFE  [fuel] = salesSum[fuel] / feSum[fuel];
                ratedFE   [fuel] = onRoadFE[fuel] / (1 - gap[fuel]);
                fuelShares[fuel] = salesSum[fuel] / salesSum.Total;
            }
        }
        void InitializeFuelPriceArrays(int year)
        {
            Estimates                     fpEstimates  = this.Settings.OperatingModes.FuelPriceEstimates;
            Parameters.ForecastDataValues forecastData = this._parameters.ForecastData;
            FuelValue energyDensity = this._parameters.FuelProperties.EnergyDensity;
            double elcRatio = energyDensity.Gasoline / energyDensity.Electricity;
            double hgnRatio = energyDensity.Gasoline / energyDensity.Hydrogen;
            double[] fuelTax, preTaxFP, retailFP;
            this.RetailFP        = new double[FuelTypeCount][];
            this.RetailFPBaseVMT = new double[FuelTypeCount];
            for (int i = 0; i < FuelTypeCount; i++)
            {   
                fuelTax  = forecastData.GetFuelTax        (             FuelTypes[i]);
                preTaxFP = forecastData.GetPreTaxFuelPrice(fpEstimates, FuelTypes[i]);
                retailFP = forecastData.GetRetailFuelPrice(fpEstimates, FuelTypes[i]);
                this.RetailFP[i] = new double[Interaction.CY];
                double edRatio = (FuelTypes[i] == FuelType.Hydrogen) ? hgnRatio : (FuelTypes[i] == FuelType.Electricity) ? elcRatio : 1;
                for (int j = 0; j < Interaction.CY; j++)
                {
                    int index = Math.Min(year + j, Parameters.ForecastDataValues.MaxYear) - Parameters.ForecastDataValues.MinYear;
                    this.RetailFP[i][j] = retailFP[index] * edRatio;
                }
                this.RetailFPBaseVMT[i] = retailFP[this._parameters.EconomicData.VMTBaseYear - Parameters.ForecastDataValues.MinYear] * edRatio;
            } 
        }
        #endregion
        #region 
        public override IModelingProgress Progress
        {
            get
            {
                bool reporting = (this.XlReportGenerator != null && this.XlReportGenerator.Reporting);
                string reportingProgress = (reporting) ? this.XlReportGenerator.ReportingProgress : null;
                if (this.Running || reporting)
                {
                    if (!this._eisRunning) { return base.Progress; }
                    else
                    {
                        string additionalInfo = "Running additional analysis for the EIS.";
                        if (reporting)
                        {
                            if (additionalInfo == null || additionalInfo == string.Empty) { additionalInfo = reportingProgress; }
                            else { additionalInfo += ("\n" + reportingProgress); }
                        }
                        return new ModelingProgress(this._scen, this._year, this._mfr, additionalInfo);
                    }
                }
                return null;
            }
        }
        public EISModelData EISData { get { return this._eisData; } }
        #endregion
        #region 
        static readonly FuelType[] FuelTypes;
        static readonly int FuelTypeCount;
        EISSettings _eisSettings;
        Parameters           _parameters;
        EISParameters        _eisParameters;
        EISTailpipeEmissions _eisTailpipeEmissions;
        int  _eisMinYear;   
        int  _eisMaxYear;   
        bool _eisRunning;   
        EISModelData _eisData;
        double[][] RetailFP;        
        double[]   RetailFPBaseVMT; 
        #endregion
    }
}

