#region << Using Directives >>
using System;
using System.Collections.Generic;
using Volpe.Cafe.Data;
using Volpe.Cafe.IO;
using Volpe.Cafe.IO.Reporting.CSV;
using Volpe.Cafe.Model;
using Volpe.Cafe.Optimization.Data;
using Volpe.Cafe.Optimization.IO.Reporting;
using Volpe.Cafe.Optimization.Settings;
using Volpe.Cafe.Settings;
#endregion
namespace Volpe.Cafe.Optimization.Model
{
    [Serializable]
    [ModelDescription("Optimization Model", "Optimizes the shape of the reformed CAFE standard by running multiple compliance " +
         "iterations and taking into account the resultant technology costs and benefits.", 1.0F)]
    public sealed class OptimizationModel : Compliance, IOptimizationModel
    {
        #region 
        public OptimizationModel() : base() { }
        #endregion
        #region 
        #region 
        public override ICompliance CreateNew()
        {
            return new OptimizationModel();
        }
        protected override CsvReportGenerator CreateCsvReportGenerator()
        {
            return new CsvOptimizationReportGenerator(this);
        }
        protected override void StartInternal()
        {
            if (this.PreProcess())
            {   
                this.Begin();
                this.PostProcess();
            }
        }
        protected override string Validate()
        {
            string message = base.Validate();
            if (message != string.Empty) { return message; }
            if (this.ScenCount != 2)
            {
                message = "The optimization model requires exactly 2 scenarios to be present in the scenarios file:  the " +
                    "Baseline scenario and the Optimization Definition scenario.  The number of scenarios found was " +
                    this.ScenCount + ".";
            }
            OptimizationSettings optSettings = (OptimizationSettings)this.Settings.GetRuntimeSettings(typeof(OptimizationSettings));
            if (optSettings == null)
            {
                message = "The modeling settings for the optimization model were not found.";
            }
            else if (optSettings.Type != OptimizationType.PassengerAuto &&
                optSettings.Type != OptimizationType.LightTruck &&
                optSettings.Type != OptimizationType.Default)
            {
                message = "The specified optimization type is not valid.  Valid types are PassengerAuto, LightTruck, or Default.";
            }
            if (message != string.Empty)
            {
                message = "The following errors have occured during initialization:\n" + message + "\n\nThe model will not run.";
            }
            return message;
        }
        protected override void OnModelingChanged()
        {
            if (!this.Running) { this.ClearStatus(); }
            base.OnModelingChanged();
        }
        #endregion
        #region 
        protected override bool PreProcess()
        {
            if (this.State == ModelingState.Completed) { return false; }
            this._optSettings = (OptimizationSettings)this.Settings.GetRuntimeSettings(typeof(OptimizationSettings));
            if (this._optSettings.Type == OptimizationType.Default)
            {
                Scenario  typeScen = this.Settings.Scenarios[1];
                bool      typeDetermined = false;
                for (int i = 0; i < this.MfrCount; i++)
                {   
                    if (this.Data.Manufacturers[i].Description.Optimize)
                    {
                        List<Vehicle> vehs = this.Data.Manufacturers[i].Vehicles;
                        for (int j = 0, vehCount = vehs.Count; j < vehCount; j++)
                        {
                            RegulatoryClass regClass = Standards.GetRegClass(vehs[j], typeScen);
                            if (regClass == RegulatoryClass.LightTruck)
                            {
                                this._optSettings.Type = OptimizationType.LightTruck;
                                typeDetermined = true;
                            }
                            else if (regClass != RegulatoryClass.Unregulated)
                            {
                                this._optSettings.Type = OptimizationType.PassengerAuto;
                                typeDetermined = true;
                            }
                            if (typeDetermined) { break; }
                        } 
                        if (typeDetermined) { break; }
                    } 
                } 
            }
            this._optMinYear = this._optSettings.StartYear;
            if (this._optMinYear < this.MinYear) { this._optMinYear = this.MinYear; }
            this._optMfrCount = 0;
            for (int i = 0; i < this.MfrCount; i++)
            {
                if (this.Data.Manufacturers[i].Description.Optimize) { this._optMfrCount++; }
            }
            this._minFP = Math.Round(this.Data.ModelData.MinFP, 1) - 0.1;
            this._maxFP = Math.Round(this.Data.ModelData.MaxFP, 1) + 0.1;
            this._stage  = OptimizationStage.None;  
            this._status = string.Empty;
            this._mfrSales = new double[this.MfrCount][];
            List<Manufacturer> mfrs = this.Data.Manufacturers;
            for (int i = 0; i < this.MfrCount; i++)
            {
                this._mfrSales[i] = new double[this.YearCount];
                List<Vehicle> vehs = mfrs[i].Vehicles;
                for (int j = 0; j < this.YearCount; j++)
                {
                    for (int k = 0, vehCount = vehs.Count; k < vehCount; k++)
                    {
                        this._mfrSales[i][j] += vehs[k].Description.Sales[this.MinYearIndex + j];
                    } 
                } 
            } 
            this._baselineData     = new IterationData [this.YearCount];
            this._optimizationData = new IterationsData[this.YearCount];
            this._iterationCount   = this._optSettings.AboveOptimum + this._optSettings.BelowOptimum + 1;
            for (int i = 0; i < this.YearCount; i++)
            {
                this._baselineData    [i] = new IterationData();
                this._optimizationData[i] = new IterationsData(this._iterationCount);
            }
            this._optimizedData = new Industry[this.YearCount];
            return base.PreProcess();
        }
        #endregion
        #region 
        void Begin()
        {
            this.ProcessBaselineScenario();
            if (this.AbortRequested) { return; }
            this.ProcessOptimizationScenario();
            if (this.AbortRequested) { return; }
            this._stage = OptimizationStage.None;
        }
        void ProcessBaselineScenario()
        {
            this._stage = OptimizationStage.Baseline;
            Industry data = this.Data.Clone();
            base.RunScenario(this.Settings.Scenarios[0], data);
            if (!this.AbortRequested)
            {   
                for (int i = this.MinYearIndex; i <= this.MaxYearIndex; i++)
                {
                    data = this.GetData(this.Settings.Scenarios[0], this.ModelYears[i - this.MinYearIndex]);    
                    this.SaveEntry(-1, i, data, null);                                                          
                }
            }
        }
        void ProcessOptimizationScenario()
        {
            Scenario scen = this.Settings.Scenarios[1];
            this._stage   = OptimizationStage.Industry;
            this._scen    = scen;  
            for (int i = this.MinYear; i <= this.MaxYear; i++)
            {   
                ModelYear year = new ModelYear(i);
                this.OptimizeModelYear(scen, year);
                if (this.AbortRequested) { break; }
            }
            this.ClearStatus();
            base.PostProcessScenario(scen);
        }
        void OptimizeModelYear(Scenario scen, ModelYear year)
        {
            if (year.Year < this._optMinYear)
            {   
                this.RunModelYearNonOptimized(scen, year);
            }
            else
            {   
                this.RunModelYearOptimized(scen, year);
            }
        }
        void RunModelYearNonOptimized(Scenario scen, ModelYear year)
        {
            int snIndex = scen.Index;
            int yrIndex = year.Index;
            this.InitializeModelYear(scen, year);
            base.RunModelYear(scen, year);
            this._optimizedData[yrIndex - this.MinYearIndex] = this.GetData(scen, year).Clone();
            if (this.Settings.OperatingModes.MultiYearModeling)
            {   
                for (int i = this.MinYearIndex; i < yrIndex; i++)
                {
                    this._optimizedData[i - this.MinYearIndex] = this.GetData(scen, this.ModelYears[i - this.MinYearIndex]).Clone();
                }
            } 
        }
        void RunModelYearOptimized(Scenario scen, ModelYear year)
        {
            this._year = year;
            int       yrYear  = year.Year;
            int       yrIndex = year.Index;
            bool      optLt   = (this._optSettings.Type == OptimizationType.LightTruck);
            int   []  fncType = (optLt) ? scen.LtFunctionType : scen.AutoFunctionType;
            double[,] fncCoef = (optLt) ? scen.LtCoefficients : scen.AutoCoefficients;
            bool restartYear = false;
            double aGpm = 0, bGpm = 0, lnrSlope = 0, lnrYGpm = 0;
            do
            {
                if (restartYear) { restartYear = false; }   
                this.PreProcessModelYearForOptimization (scen, year, fncType, fncCoef, ref aGpm, ref bGpm, ref lnrSlope, ref lnrYGpm);
                this.IterationLoop                      (scen, year, fncType, fncCoef,     aGpm,     bGpm,     lnrSlope,     lnrYGpm);
                this.PostProcessModelYearForOptimization(scen, year, fncType, fncCoef, ref restartYear);
            } while (restartYear);
            if (this.AbortRequested) { return; }
            this.SetStatus("Modeling for optimized MY-" + yrYear + " fleet ...");
            this.WriteLogLine("Modeling for optimized MY-" + yrYear + " fleet ...");
            this.WriteLogLine(CharAsterisk50);
            double[] optCoefs = this._optimizationData[yrYear - this.MinYear].GetCoefficients();
            for (int i = 0; i < 4; i++) { fncCoef[i, yrIndex] = optCoefs[i]; }
            this.RunModelYearNonOptimized(scen, year);
        }
        void PreProcessModelYearForOptimization(Scenario scen, ModelYear year, int[] fncType, double[,] fncCoef,
            ref double aGpm, ref double bGpm, ref double lnrSlope, ref double lnrYGpm)
        {
            int yrIndex = year.Index;
            if (FunctionInformation.IsLinearFunction(fncType[yrIndex]))
            {   
                double coefIncr = Math.Pow(1 - this._optSettings.Increment, -this._optSettings.AboveOptimum);
                aGpm     = 1 / fncCoef[0, yrIndex] * coefIncr;
                bGpm     = 1 / fncCoef[1, yrIndex] * coefIncr;
                lnrSlope =     fncCoef[2, yrIndex] * coefIncr;
                lnrYGpm  =     fncCoef[3, yrIndex] * coefIncr;
            }
            else
            {   
                double coefIncr = this._optSettings.Increment * this._optSettings.AboveOptimum;
                aGpm = 1 / fncCoef[0, yrIndex] + coefIncr;
                bGpm = 1 / fncCoef[1, yrIndex] + coefIncr;
            }
        }
        void PostProcessModelYearForOptimization(Scenario scen, ModelYear year, int[] fncType, double[,] fncCoef,
            ref bool restartYear)
        {
            if (this.AbortRequested) { return; }
            int  yrIndex  = year.Index;
            int  yrOffset = yrIndex - this.MinYearIndex;
            int  currType = fncType[yrIndex];
            bool isLinear = FunctionInformation.IsLinearFunction(currType);
            this.CalcOptimizedValues(yrIndex);
            if (this._optSettings.BackstopAsymptotes && yrIndex > this.MinYearIndex && currType == fncType[yrIndex - 1])
            {   
                double[] currCoef = this._optimizationData[yrOffset].GetCoefficients();
                double[] prevCoef = new double[4];
                for (int i = 0; i < 4; i++) { prevCoef[i] = fncCoef[i, yrIndex - 1]; }
                bool boundsDecreased =
                    (Math.Round(currCoef[0], 1) < Math.Round(prevCoef[0], 1)) ||
                    (Math.Round(currCoef[1], 1) < Math.Round(prevCoef[1], 1));
                if (boundsDecreased)
                {   
                    this.SetStatus("Backstopping Asymptotes for MY-" + year.Year + " ...");
                    this.WriteLogLine("Backstopping Asymptotes for MY-" + year.Year + " ...");
                    this.WriteLogLine(CharAsterisk50);
                    if (Math.Round(currCoef[0], 1) < Math.Round(prevCoef[0], 1)) { currCoef[0] = prevCoef[0]; }
                    if (Math.Round(currCoef[1], 1) < Math.Round(prevCoef[1], 1)) { currCoef[1] = prevCoef[1]; }
                }
            } 
            if (this._optSettings.BackstopTargetFunction && yrIndex > this.MinYearIndex && currType == fncType[yrIndex - 1])
            {   
                double[] currCoef = this._optimizationData[yrOffset].GetCoefficients();
                double[] prevCoef = new double[4];
                for (int i = 0; i < 4; i++) { prevCoef[i] = fncCoef[i, yrIndex - 1]; }
                bool boundsDecreased =
                    (Math.Round(currCoef[0], 1) < Math.Round(prevCoef[0], 1)) ||
                    (Math.Round(currCoef[1], 1) < Math.Round(prevCoef[1], 1)) ||
                    this.CheckCurvesCrossing(currCoef, prevCoef, currType, this._minFP, this._maxFP, 0.1);
                if (boundsDecreased && (isLinear || (currCoef[2] == prevCoef[2] && currCoef[3] == prevCoef[3])))
                {   
                    this.SetStatus("Backstopping Target Function for MY-" + year.Year + " asymptotes ...");
                    this.WriteLogLine("Backstopping Target Function for MY-" + year.Year + " asymptotes ...");
                    this.WriteLogLine(CharAsterisk50);
                    this._optimizationData[yrOffset].SetOptimizedIndexFromCoefs(Math.Round(prevCoef[0], 1),
                        Math.Round(prevCoef[1], 1));
                }
                else if (boundsDecreased)
                {   
                    this._optimizationData[yrOffset].Clear();
                    double shapeOff = this._optSettings.Increment * (this._optimizationData[yrOffset - 1].Optimized -
                        this._optSettings.AboveOptimum);
                    fncCoef[0, yrIndex] = 1 / (1 / prevCoef[0] + shapeOff);
                    fncCoef[1, yrIndex] = 1 / (1 / prevCoef[1] + shapeOff);
                    fncCoef[2, yrIndex] = prevCoef[2];
                    fncCoef[3, yrIndex] = prevCoef[3];
                    restartYear = true;
                    this.SetStatus("Backstopping Target Function for MY-" + year.Year + " shape (a restart will be required) ...");
                    this.WriteLogLine("Backstopping Target Function for MY-" + year.Year + " shape (a restart will be required) ...");
                    this.WriteLogLine(CharAsterisk50);
                    return;
                }
            } 
        }
        void IterationLoop(Scenario scen, ModelYear year, int[] fncType, double[,] fncCoef, double aGpm, double bGpm,
            double lnrSlope, double lnrYGpm)
        {
            int cfExhaustedCount = 0;
            bool cfExhausted;
            for (int i = 0; i < this._iterationCount; i++)
            {
                this.PreProcessIteration(i, scen, year, fncType, fncCoef, aGpm, bGpm, lnrSlope, lnrYGpm);
                base.RunModelYear(scen, year);
                this.PostProcessIteration(i, scen, year, fncType, fncCoef, out cfExhausted);
                if (cfExhausted) { cfExhaustedCount++; }
                if (cfExhaustedCount == 2) { return; }
                if (this.AbortRequested) { return; }
            }
        }
        void PreProcessIteration(int iter, Scenario scen, ModelYear year, int[] fncType, double[,] fncCoef, double aGpm,
            double bGpm, double lnrSlope, double lnrYGpm)
        {
            int yrIndex = year.Index;
            this.SetStatus("Examining iteration: " + (iter + 1) + " / " + this._iterationCount);
            this.InitializeModelYear(scen, year);
            bool   isLinear  = FunctionInformation.IsLinearFunction(fncType[yrIndex]);
            if (FunctionInformation.IsLinearFunction(fncType[yrIndex]))
            {   
                double coefIncr = Math.Pow(1 - this._optSettings.Increment, iter);
                fncCoef[0, yrIndex] = Math.Round(1 / (aGpm * coefIncr), 1);
                fncCoef[1, yrIndex] = Math.Round(1 / (bGpm * coefIncr), 1);
                fncCoef[2, yrIndex] = lnrSlope * coefIncr;
                fncCoef[3, yrIndex] = lnrYGpm  * coefIncr;
            }
            else
            {   
                double coefDelta = iter * this._optSettings.Increment;
                fncCoef[0, yrIndex] = Math.Round(1 / (aGpm - coefDelta), 1);
                fncCoef[1, yrIndex] = Math.Round(1 / (bGpm - coefDelta), 1);
            }
        }
        void PostProcessIteration(int iter, Scenario scen, ModelYear year, int[] fncType, double[,] fncCoef, out bool cfExhausted)
        {
            if (this.AbortRequested) { cfExhausted = false; return; }
            int yrIndex = year.Index;
            Industry data = this.GetData(scen, year);
            int coefCount = FunctionInformation.GetCoefficientCount(fncType[yrIndex]);
            double[] coefs = new double[coefCount];
            for (int i = 0; i < coefCount; i++) { coefs[i] = fncCoef[i, yrIndex]; }
            this.SaveEntry(iter, yrIndex, data, coefs);
            cfExhausted = true;
            List<Manufacturer> mfrs = data.Manufacturers;
            for (int i = 0; i < this.MfrCount; i++)
            {   
                if (mfrs[i].Description.Optimize)
                {   
                    if (!mfrs[i].ModelData.CFExhausted) { cfExhausted = false; break; }
                }
            }
            if (iter + 1 == this._iterationCount) { this.WriteLogLine(CharAsterisk50); }
            else                                  { this.WriteLogLine(CharDash50    ); }
        }
        void InitializeModelYear(Scenario scen, ModelYear year)
        {
            if (year.Index == this.MinYearIndex)
            {   
                Industry data = this.Data.Clone();
                base.PreProcessScenario(scen, data);
                this.SetActiveScenarioData(year, data);
            }
            else
            {   
                if (this.Settings.OperatingModes.MultiYearModeling)
                {   
                    for (int i = this.MinYearIndex; i < year.Index; i++)
                    {
                        Industry optData = this._optimizedData[i - this.MinYearIndex].Clone();
                        this.SetActiveScenarioData(this.ModelYears[i - this.MinYearIndex], optData);
                    }
                } 
            }
        }
        void SaveEntry(int iteration, int yrIndex, Industry data, double[] coefficients)
        {
            bool optAuto = (this._optSettings.Type == OptimizationType.PassengerAuto);
            bool optLt   = (this._optSettings.Type == OptimizationType.LightTruck);
            if (!optAuto && !optLt) { return; }
            int yrOffset = yrIndex - this.MinYearIndex;     
            if      (this._stage == OptimizationStage.Baseline) { this._baselineData[yrOffset].Reset(); }
            else if (this._stage == OptimizationStage.Industry) { this._optimizationData[yrOffset].Iterations[iteration].Reset(); }
            double salesSum = 0, stndSum = 0, cafeSum = 0;
            double complianceCosts = 0;
            double socialCosts     = 0;
            double discSocialCosts = 0;
            double fuelUse         = 0;
            for (int i = 0, mfrCount = data.ManufacturerCount; i < mfrCount; i++)
            {
                Manufacturer mfr = data.Manufacturers[i];
                if (!mfr.Description.Optimize) { continue; }
                Manufacturer.CModelData mmd = mfr.ModelData;
                EffectsData             med = mfr.EffectsData;
                if (optLt)
                {   
                    double ltSales = mmd.Sales   .LightTruck; if (double.IsNaN(ltSales)) { ltSales = 0; }
                    double ltStnd  = mmd.Standard.LightTruck; if (double.IsNaN(ltSales)) { ltSales = 0; }
                    double ltCAFE  = mmd.CAFE    .LightTruck; if (double.IsNaN(ltSales)) { ltSales = 0; }
                    salesSum += ltSales;
                    stndSum  += (ltStnd == 0) ? 0 : ltSales / ltStnd;
                    cafeSum  += (ltCAFE == 0) ? 0 : ltSales / ltCAFE;
                    complianceCosts += mmd.TotalSocialCosts                                   .LightTruck / 1e6;
                    socialCosts     += med.UndiscountedOwnerAndSocietalCosts .TotalSocialCosts.LightTruck / 1e3;
                    discSocialCosts += med.DiscountedOwnerAndSocietalCosts[0].TotalSocialCosts.LightTruck / 1e3;
                    fuelUse         += med.Gallons                                            .LightTruck.Total / 1e3;
                }
                else
                {   
                    double daSales = mmd.Sales.DomesticAuto;        if (double.IsNaN(daSales)) { daSales = 0; }
                    double iaSales = mmd.Sales.ImportedAuto;        if (double.IsNaN(iaSales)) { iaSales = 0; }
                    double daStnd  = mmd.Standard.DomesticAuto;     if (double.IsNaN(daStnd )) { daStnd  = 0; }
                    double iaStnd  = mmd.Standard.ImportedAuto;     if (double.IsNaN(iaStnd )) { iaStnd  = 0; }
                    double daCAFE  = mmd.CAFE.DomesticAuto;         if (double.IsNaN(daCAFE )) { daCAFE  = 0; }
                    double iaCAFE  = mmd.CAFE.ImportedAuto;         if (double.IsNaN(iaCAFE )) { iaCAFE  = 0; }
                    salesSum = daSales + iaSales;
                    stndSum  = ((daStnd == 0) ? 0 : daSales / daStnd) + ((iaStnd == 0) ? 0 : iaSales / iaStnd);
                    cafeSum  = ((daCAFE == 0) ? 0 : daSales / daCAFE) + ((iaCAFE == 0) ? 0 : iaSales / iaCAFE);
                    complianceCosts += mmd.TotalSocialCosts                                   .PassengerAuto / 1e6;
                    socialCosts     += med.UndiscountedOwnerAndSocietalCosts .TotalSocialCosts.PassengerAuto / 1e3;
                    discSocialCosts += med.DiscountedOwnerAndSocietalCosts[0].TotalSocialCosts.PassengerAuto / 1e3;
                    fuelUse         += med.Gallons                                            .PassengerAuto.Total / 1e3;
                }
            } 
            if (this._stage == OptimizationStage.Baseline)
            {   
                this._baselineData[yrOffset].Valid    = true;
                this._baselineData[yrOffset].Standard = salesSum / stndSum;
                this._baselineData[yrOffset].CAFE     = salesSum / cafeSum;
                this._baselineData[yrOffset].TotalSocialCosts        = complianceCosts;
                this._baselineData[yrOffset].TotalSocialBenefits     = socialCosts;
                this._baselineData[yrOffset].TotalSocialBenefitsDisc = discSocialCosts;
                this._baselineData[yrOffset].FuelSavings             = fuelUse;
            }
            else if (this._stage == OptimizationStage.Industry)
            {   
                IterationData iterData = this._optimizationData[yrOffset].Iterations[iteration];
                iterData.Valid        = true;
                iterData.Coefficients = coefficients;
                iterData.Standard     = salesSum / stndSum;
                iterData.CAFE         = salesSum / cafeSum;
                iterData.TotalSocialCosts        = -(this._baselineData[yrOffset].TotalSocialCosts        - complianceCosts);
                iterData.TotalSocialBenefits     =  (this._baselineData[yrOffset].TotalSocialBenefits     - socialCosts    );
                iterData.TotalSocialBenefitsDisc =  (this._baselineData[yrOffset].TotalSocialBenefitsDisc - discSocialCosts);
                iterData.FuelSavings             =  (this._baselineData[yrOffset].FuelSavings             - fuelUse        );
                iterData.NetBenefits             =  (iterData.TotalSocialBenefitsDisc                     - iterData.TotalSocialCosts);
            }
        }
        void CalcOptimizedValues(int yrIndex)
        {
            int yrOffset = yrIndex - this.MinYearIndex;
            IterationData[] iterations = this._optimizationData[yrOffset].Iterations;
            int optIdx = 0;
            if      (this._optSettings.Mode == OptimizationMode.MaximumNetBenefits) { optIdx = this.GetMaxNetBenefits (iterations); }
            else if (this._optSettings.Mode == OptimizationMode.ZeroNetBeneftis   ) { optIdx = this.GetZeroNetBenefits(iterations); }
            this.SaveOptimizedData(yrOffset, optIdx, this._optimizationData);
        }
        int GetMaxNetBenefits(IterationData[] iterations)
        {
            int max = 0;    
            for (int i = 0, count = iterations.Length; i < count; i++)
            {
                if (iterations[i].Valid)
                {   
                    if (iterations[max].NetBenefits < iterations[i].NetBenefits)
                    {   
                        max = i;
                    }
                }
            } 
            return (iterations[max].Valid) ? max : -1;    
        }
        int GetZeroNetBenefits(IterationData[] iterations)
        {
            int max = this.GetMaxNetBenefits(iterations);
            if (max == - 1) { return -1; }
            int min = max, lastValid = max;
            for (int i = max, count = iterations.Length; i < count; i++)
            {
                if (iterations[i].Valid)
                {   
                    if (iterations[i].NetBenefits < 0) { break; }
                    if (iterations[i].NetBenefits >= 0 && iterations[i].NetBenefits < iterations[min].NetBenefits)
                    {   
                        min = i;
                    }
                }
            } 
            return (iterations[min].Valid) ? min : -1;    
        }
        void SaveOptimizedData(int yrOffset, int optIdx, IterationsData[] data)
        {
            if      (optIdx == -1 && yrOffset > 0) { data[yrOffset].Optimized = data[yrOffset - 1].Optimized; }
            else if (optIdx == -1                ) { data[yrOffset].Optimized = -1; }
            else if (this._optSettings.BackstopStandard && yrOffset > 0 && data[yrOffset - 1].Optimized >= 0 &&
                data[yrOffset].Iterations[optIdx].Standard < data[yrOffset - 1].GetStandard())
            {
                data[yrOffset].SetOptimizedIndexFromStnd(data[yrOffset - 1].GetStandard());
            }
            else { data[yrOffset].Optimized = optIdx; }
        }
        bool CheckCurvesCrossing(double[] currCoef, double[] prevCoef, int fncType, double fpStart, double fpEnd, double fpIncrement)
        {
            for (double fp = fpStart; fp < fpEnd; fp += fpIncrement)
            {   
                double currTarget = Standards.GetTarget(fp, 0, fncType, currCoef);
                double prevTarget = Standards.GetTarget(fp, 0, fncType, prevCoef);
                if (currTarget > prevTarget) { return true; }
            }
            return false;
        }
        void SetStatus(string status)
        {
            this._status = status;
        }
        void ClearStatus()
        {
            this._status = string.Empty;
        }
        void WriteLogLine(string value)
        {
            ComplianceWriter log    = (this.LogWriter.CFs    == null) ? null : this.LogWriter.CFs   [this._scen.Index];
            ComplianceWriter extLog = (this.LogWriter.ExtCFs == null) ? null : this.LogWriter.ExtCFs[this._scen.Index];
            for (int i = 0; i < this.MfrCount; i++)
            {
                if (log    != null) { log   .Write(value); }
                if (extLog != null) { extLog.Write(value); }
            }
        }
        #endregion
        #endregion
        #region 
        #region 
        public override IModelingProgress Progress
        {
            get
            {
                bool   xlReporting          = (this.XlReportGenerator  != null && this.XlReportGenerator .Reporting);
                bool   csvReporting         = (this.CsvReportGenerator != null && this.CsvReportGenerator.Reporting);
                bool   reporting            =  xlReporting || csvReporting;
                string xlReportingProgress  = (xlReporting) ? this.XlReportGenerator.ReportingProgress : null;
                string csvReportingProgress = (csvReporting) ? this.CsvReportGenerator.ReportingProgress : null;
                string reportingProgress    =  xlReportingProgress + ((xlReporting) ? "" : "\n") + csvReportingProgress;
                if (this.Running || reporting)
                {
                    if      (this._stage == OptimizationStage.Baseline) { return base.Progress; }
                    else if (this._stage == OptimizationStage.Industry)
                    {
                        string additionalInfo = this._status;
                        if (reporting)
                        {
                            if (additionalInfo == null || additionalInfo == string.Empty) { additionalInfo = reportingProgress; }
                            else { additionalInfo += ("\n" + reportingProgress); }
                        }
                        return new ModelingProgress(this._scen, this._year, this._mfr, additionalInfo);
                    }
                }
                return null;
            }
        }
        #endregion
        public OptimizationData OptimizationData { get { return new OptimizationData(this._optimizationData); } }
        #endregion
        #region 
        const string CharAsterisk50 = "**************************************************";
        const string CharDash50     = "--------------------------------------------------";
        OptimizationSettings _optSettings;
        int _optMinYear;
        int _optMfrCount;
        double _minFP, _maxFP;
        OptimizationStage _stage;
        string _status;
        double[][] _mfrSales;                   
        IterationData [] _baselineData;         
        IterationsData[] _optimizationData;     
        int              _iterationCount;
        Industry[] _optimizedData;
        #endregion
    }
}

