using System;
using Volpe.Cafe.Settings;
namespace Volpe.Cafe.Data
{
    [Serializable]
    public sealed class KMYType
    {
        #region 
        KMYType() { }
        KMYType(ModelYear year, Parameters parameters, Estimates priceEstimates, Manufacturer mfr, bool useOCPayback)
        {
            this._year = year;
            this._gas = this._dsl = this._e85 = this._elc = this._hgn = 0;
            Parameters.VehicleAgeDataValues vehAgeData   = parameters.VehicleAgeData;
            Parameters.ForecastDataValues   forecastData = parameters.ForecastData;
            Parameters.EconomicDataValues   econData     = parameters.EconomicData;
            double discRate      = mfr.Description.DiscountRate;
            double paybackPeriod = (useOCPayback) ? mfr.Description.PaybackPeriod_OC : mfr.Description.PaybackPeriod;
            int    payback       = (int)Math.Floor(paybackPeriod);
            double fract         = paybackPeriod - payback;         
            int    yrYear        = this._year.Year;
            double[] gasPriceArr = forecastData.GetRetailFuelPrice(priceEstimates, FuelType.Gasoline   );
            double[] dslPriceArr = forecastData.GetRetailFuelPrice(priceEstimates, FuelType.Diesel     );
            double[] e85PriceArr = forecastData.GetRetailFuelPrice(priceEstimates, FuelType.Ethanol85  );
            double[] elcPriceArr = forecastData.GetRetailFuelPrice(priceEstimates, FuelType.Electricity);
            double[] hgnPriceArr = forecastData.GetRetailFuelPrice(priceEstimates, FuelType.Hydrogen   );
            double gasPrice = 0, dslPrice = 0, e85Price = 0, elcPrice = 0, hgnPrice = 0;
            FuelValue energyDensity = parameters.FuelProperties.EnergyDensity;
            double elcRatio = energyDensity.Gasoline / energyDensity.Electricity;
            double hgnRatio = energyDensity.Gasoline / energyDensity.Hydrogen;
            for (int i = 0; i < payback; i++)
            {   
                double surv = (vehAgeData.SurvivalCars[i] + vehAgeData.SurvivalVans[i] + vehAgeData.SurvivalSUVs[i] + vehAgeData.SurvivalPickups[i]) / 4;
                double vmtGrowth = econData.GetVMTGrowthFactor(priceEstimates, yrYear, i);
                double miles = (vehAgeData.MilesCars[i] + vehAgeData.MilesVans[i] + vehAgeData.MilesSUVs[i] + vehAgeData.MilesPickups[i]) * vmtGrowth / 4;
                int yrIndex = Math.Min(year.Year + i, Parameters.ForecastDataValues.MaxYear) - Parameters.ForecastDataValues.MinYear;
                double multiplier = surv * miles / Math.Pow(1 + discRate, i);
                gasPrice = gasPriceArr[yrIndex] * multiplier;
                dslPrice = dslPriceArr[yrIndex] * multiplier;
                e85Price = e85PriceArr[yrIndex] * multiplier;
                elcPrice = elcPriceArr[yrIndex] * multiplier * elcRatio;
                hgnPrice = hgnPriceArr[yrIndex] * multiplier * hgnRatio;
                this._gas += gasPrice;
                this._dsl += dslPrice;
                this._e85 += e85Price;
                this._elc += elcPrice;
                this._hgn += hgnPrice;
            }
            if (fract > 0D)
            {   
                this._gas += (fract * gasPrice);
                this._dsl += (fract * dslPrice);
                this._e85 += (fract * e85Price);
                this._elc += (fract * elcPrice);
                this._hgn += (fract * hgnPrice);
            }
        }
        #endregion
        #region 
        internal KMYType Clone()
        {
            KMYType kmy = new KMYType();
            kmy._year = new ModelYear(this._year.Year);
            kmy._gas  = this._gas;
            kmy._dsl  = this._dsl;
            kmy._e85  = this._e85;
            kmy._elc  = this._elc;
            kmy._hgn  = this._hgn;
            return kmy;
        }
        public override string ToString()
        {
            return "{ModelYear=" + this._year.ToString() +
                ", Gasoline="    + this._gas.ToString() +
                ", Diesel="      + this._dsl.ToString() +
                ", Ethanol-85="  + this._e85.ToString() +
                ", Electricity=" + this._elc.ToString() +
                ", Hydrogen="    + this._hgn.ToString() + "}";
        }
        internal static KMYType[] GetKmyValues(Parameters parameters, Estimates priceEstimates, Manufacturer mfr, bool useOCPayback)
        {
            int numYears = mfr.Description.WillingToPayFines.Length;
            KMYType[] kmy = new KMYType[numYears];
            for (int i = 0; i < numYears; i++)
            {
                kmy[i] = new KMYType(new ModelYear(ModelYear.FromIndex(i)), parameters, priceEstimates, mfr, useOCPayback);
            }
            return kmy;
        }
        #endregion
        #region 
        public double this[FuelType fuelType]
        {
            get
            {
                return
                    (fuelType == FuelType.Gasoline   ) ? this._gas :
                    (fuelType == FuelType.Diesel     ) ? this._dsl :
                    (fuelType == FuelType.Ethanol85  ) ? this._e85 :
                    (fuelType == FuelType.Electricity) ? this._elc :
                    (fuelType == FuelType.Hydrogen   ) ? this._hgn : 0;
            }
        }
        public ModelYear ModelYear { get { return this._year; } }
        public double Gasoline { get { return this._gas; } }
        public double Diesel { get { return this._dsl; } }
        public double Ethanol85 { get { return this._e85; } }
        public double Electricity { get { return this._elc; } }
        public double Hydrogen { get { return this._hgn; } }
        #endregion
        #region 
        ModelYear _year;
        double _gas, _dsl, _e85, _elc, _hgn;
        #endregion
    }
}

