#region << Using Directives >>
using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Threading;
using Volpe.Cafe;
using Volpe.Cafe.Data;
using Volpe.Cafe.IO;
using Volpe.Cafe.Settings;
using TI = Volpe.Cafe.TechnologyIndexes;
#endregion
namespace Volpe.Cafe.Model
{
    [Serializable]
    [ModelDescription("Standard Compliance Model", "Runs a regular compliance model that estimates technology costs and " +
         "benefits under scenarios with predefined flat or reformed CAFE standards.", 1.0F)]
    public class Compliance : ComplianceBase
    {
        #region 
        public Compliance() : base() { }
        #endregion
        #region 
        #region 
        public override ICompliance CreateNew()
        {
            return new Compliance();
        }
        protected override void StartInternal()
        {
            if (this.PreProcess())
            {   
                this.ScenarioLoop();
                this.PostProcess();
            }
        }
        #endregion
        #region 
        protected virtual bool PreProcess()
        {
            if (this.State == ModelingState.Completed) { return false; }
            for (int i = 0; i < this.MfrCount; i++)
            {
                Manufacturer              mfr = this.Data.Manufacturers[i];
                Manufacturer.CDescription md  = mfr.Description;
                if (md.DiscountRate < 0)
                {   
                    md.DiscountRate = this.Settings.Parameters.EconomicData.PrivateDiscountRate;
                }
                Estimates priceEstimates = this.Settings.OperatingModes.FuelPriceEstimates;
                mfr.ModelData.KMY    = KMYType.GetKmyValues(this.Settings.Parameters, priceEstimates, mfr, false);
                mfr.ModelData.KMY_OC = KMYType.GetKmyValues(this.Settings.Parameters, priceEstimates, mfr, true);
            }
            this._scen = null;
            this._year = null;
            this._mfr  = null;
            this._effectsModel = new Effects();
            return true;
        }
        protected virtual void PostProcess()
        {
        }
        #endregion
        #region 
        #region 
        protected virtual void ScenarioLoop()
        {
            for (int i = 0; i < this.ScenCount; i++)
            {   
                Scenario scen = this.Settings.Scenarios[i];
                Industry data = this.Data.Clone();
                this.RunScenario(scen, data);
                if (this.AbortRequested) { break; }
            } 
        }
        protected virtual void RunScenario(Scenario scen, Industry data)
        {
            this.PreProcessScenario (scen, data);
            this.ModelYearLoop      (scen, data);
            this.PostProcessScenario(scen);
        }
        protected virtual void PreProcessScenario(Scenario scen, Industry data)
        {
            this._scen = scen;
            this.ResetActiveScenarioData();
            this.OnScenarioStarted();
            this._lcf = new LinearComplianceFinder(scen, this.MinYear, this.MaxYear, this.Settings, this.LogWriter);
            data.ModelData.TechUsedSales = new RCDouble[TI.TechnologyCount];
            for (int i = 0; i < this.MfrCount; i++)
            {
                Manufacturer            mfr = data.Manufacturers[i];
                Manufacturer.CModelData mmd = mfr.ModelData;
                mmd.TechUsedSales     = new RCDouble [TI.TechnologyCount];
                mmd.TechAppliedSales  = new RCDouble [TI.TechnologyCount];
                mmd.TechExhausted     = new RCBoolean[TI.TechnologyCount];
                mmd.TechPhaseInOffset = new RCInt32  [TI.TechnologyCount];
                List<Vehicle> vc = mfr.Vehicles;
                for (int j = 0, vcCount = vc.Count; j < vcCount; j++)
                {   
                    vc[j].RegClass = Standards.GetRegClass(vc[j], scen);        
                    TechnologyApplicability.Initialize(vc[j], this.Settings);   
                }
            }
        }
        protected virtual void PostProcessScenario(Scenario scen)
        {
            if (this.AbortRequested) { return; }
            this._lcf = null;
            this.OnScenarioCompleted();
        }
        #endregion
        #region 
        protected virtual void ModelYearLoop(Scenario scen, Industry data)
        {
            this.SetActiveScenarioData(this.ModelYears[0], data);
            for (int i = 0; i < this.YearCount; i++)
            {   
                ModelYear year = this.ModelYears[i];
                this.RunModelYear(scen, year);
                if (this.AbortRequested) { break; }
            } 
        }
        protected virtual void RunModelYear(Scenario scen, ModelYear year)
        {
            this.PreProcessModelYear (scen, year);
            this.ManufacturerLoop    (scen, year);
            this.PostProcessModelYear(scen, year);
        }
        protected virtual void PreProcessModelYear(Scenario scen, ModelYear year)
        {
            this._year = year;
            this.OnModelYearStarted();
            int yrIndex = year.Index;
            Industry prevData = null, data = null;
            if (yrIndex != this.MinYearIndex)
            {   
                prevData = this.GetData(scen, this.ModelYears[yrIndex - this.MinYearIndex - 1]);
                this.SetActiveScenarioData(year, prevData.Clone());
            }
            data = this.GetData(scen, year);
            if (this.Settings.OperatingModes.DynamicFleetShare)
            {
                DynamicFleetShare.AdjustFleetShare(data, prevData, this.Settings, scen, year);
            }
            Industry.CModelData imd = data.ModelData;
            for (int i = 0; i < TI.TechnologyCount; i++) { imd.TechUsedSales[i].Clear(); }
            List<Technology> technologies = this.Settings.Technologies;
            for (int i = 0; i < this.MfrCount; i++)
            {   
                Manufacturer            mfr = data.Manufacturers[i];
                Manufacturer.CModelData mmd = mfr.ModelData;
                mmd.PreliminaryStandard.Clear();
                mmd.Standard           .Clear();
                mmd.TCreditsInCapped   .Clear();
                mmd.TCreditsIn         .Clear();
                mmd.TCreditsOut        .Clear();
                mmd.TechCost           .Clear();
                mmd.RegCost            .Clear();
                mmd.LossOfValue        .Clear();
                mmd.RelativeLossOfValue.Clear();
                mmd.MaintenanceCost    .Clear();
                mmd.RepairCost         .Clear();
                mmd.TaxesAndFees       .Clear();
                mmd.FinancingCost      .Clear();
                mmd.InsuranceCost      .Clear();
                for (int j = 0; j < TI.TechnologyCount; j++)
                {   
                    mmd.TechUsedSales   [j].Clear();
                    mmd.TechAppliedSales[j].Clear();
                    mmd.TechExhausted   [j].Clear();
                }
                List<Vehicle> mfrVehs = mfr.Vehicles;
                for (int j = 0, mfrVehCount = mfrVehs.Count; j < mfrVehCount; j++)
                {   
                    Vehicle              veh = mfrVehs[j];
                    Vehicle.CDescription vd  = veh.Description;
                    Vehicle.CModelData   vmd = veh.ModelData;
                    RegulatoryClass regClass = veh.RegClass;
                    TechnologyApplicability.Enable(veh, year, -1, this.Settings);
                    for (int k = 0; k < TI.TechnologyCount; k++)
                    {
                        if (vmd.TechUsed[k])
                        {   
                            imd.TechUsedSales[k][regClass] += vd.Sales[yrIndex];
                            mmd.TechUsedSales[k][regClass] += vd.Sales[yrIndex];
                            if (vmd.TechApplied[k])
                            {   
                                mmd.TechAppliedSales[k][regClass] += vd.Sales[yrIndex];
                            }
                        }
                    } 
                } 
                if (this.Settings.OperatingModes.PhaseInMode != PhaseInMode.Legacy &&
                    (yrIndex == this.MinYearIndex || this.Settings.OperatingModes.PhaseInMode == PhaseInMode.DynamicSlidingPhaseIn))
                {
                    TechnologyApplicability.ComputeMfrAdjustedPhaseIn(mfr, year, this.MinYearIndex, this.Settings);
                } 
            } 
            RCDouble indStandard = RCDouble.Zero;
            if (scen.AutoMinStndPct[yrIndex] > 0 || scen.LtMinStndPct[yrIndex] > 0)
            {
                indStandard = Standards.GetIndustryStandard(data, scen, year, false);
            }
            for (int i = 0; i < this.MfrCount; i++)
            {   
                Manufacturer            mfr     = data.Manufacturers[i];
                Manufacturer.CModelData mmd     = mfr .ModelData;
                List<Vehicle>           mfrVehs = mfr .Vehicles;
                for (int j = 0, mfrVehCount = mfrVehs.Count; j < mfrVehCount; j++)
                {   
                    Vehicle              veh           = mfrVehs[j];
                    Vehicle.CDescription vd            = veh.Description;
                    Vehicle.CModelData   vmd           = veh.ModelData;
                    double               vehCost       = 0;
                    double               vehMaintCost  = 0;
                    double               vehRepairCost = 0;
                    double               vehLossValue  = 0;
                    TechnologyClass      techClass     = veh.TechnologyClass;
                    for (int k = 0; k < TI.TechnologyCount; k++)
                    {
                        Technology tech = technologies[k];
                        if (vmd.TechUsed[k] && vmd.TechApplied[k])
                        {   
                            vehCost       += TechnologyApplicability.GetCost           (tech, veh, data, year, this.MinYear, this.MaxYear);
                            vehMaintCost  += TechnologyApplicability.GetMaintenanceCost(tech, veh,       year);
                            vehRepairCost += TechnologyApplicability.GetRepairCost     (tech, veh,       year);
                            vehLossValue += tech.Attributes[techClass].LossOfValue;
                        }
                    } 
                    if (scen.IncludeAC[year.Index])
                    {
                        vehCost += ((veh.IsRCAuto) ? scen.PCACCost[year.Index] : scen.LTACCost[year.Index]);
                    }
                    double vehSales = vd.Sales[yrIndex];
                    vmd.TechCost                       =             vehCost       ;
                    vmd.LossOfValue                    =             vehLossValue  ;
                    vmd.MaintenanceCost                =             vehMaintCost  ;
                    vmd.RepairCost                     =             vehRepairCost ;
                    mmd.TechCost       [veh.RegClass] += (vehSales * vehCost      );
                    mmd.LossOfValue    [veh.RegClass] += (vehSales * vehLossValue );
                    mmd.MaintenanceCost[veh.RegClass] += (vehSales * vehMaintCost );
                    mmd.RepairCost     [veh.RegClass] += (vehSales * vehRepairCost);
                } 
                mmd.AverageStandard = indStandard;                      
                mmd.Sales           = Standards.GetSales(mfr, year);    
                if (!this.Settings.OperatingModes.IgnorePhaseIn)
                {   
                    for (int j = 0; j < TI.TechnologyCount; j++)
                    {
                        TechnologyApplicability.CheckPhaseIn(this.Settings, mmd, RegulatoryClass.DomesticAuto, technologies[j], year, this.MinYearIndex);
                        TechnologyApplicability.CheckPhaseIn(this.Settings, mmd, RegulatoryClass.ImportedAuto, technologies[j], year, this.MinYearIndex);
                        TechnologyApplicability.CheckPhaseIn(this.Settings, mmd, RegulatoryClass.LightTruck  , technologies[j], year, this.MinYearIndex);
                        TechnologyApplicability.CheckPhaseIn(this.Settings, mmd, RegulatoryClass.Unregulated , technologies[j], year, this.MinYearIndex);
                    } 
                }
            } 
        }
        protected virtual void PostProcessModelYear(Scenario scen, ModelYear year)
        {
            if (this.AbortRequested) { return; }
            int      techCount = TI.TechnologyCount;
            Industry data      = this.GetData(scen, year);
            Industry bData     = (scen.Index == 0) ? data : this.GetData(this.Settings.Scenarios[0], year);
            Standards.CalcVehicleTaxCredit(data, scen, year);
            MarketSimulation.AllocateCosts(data, this.Settings, scen, year);
            Parameters.EconomicDataValues econData = this.Settings.Parameters.EconomicData;
            for (int i = 0, mfrCount = this.MfrCount; i < mfrCount; i++)
            {
                Manufacturer            mfr  = data.Manufacturers[i];
                Manufacturer.CModelData mmd  = mfr .ModelData;
                List<Vehicle>           vehs = mfr .Vehicles;
                mmd.RelativeLossOfValue.Clear();
                mmd.TaxesAndFees       .Clear();
                mmd.FinancingCost      .Clear();
                mmd.InsuranceCost      .Clear();
                for (int j = 0, vehCount = vehs.Count; j < vehCount; j++)
                {
                    Vehicle.CModelData vmd      = vehs[j].ModelData;
                    double             vehMsrp  = vehs[j].Description.Msrp [year.Index] + vmd.RegCost;
                    double             vehSales = vehs[j].Description.Sales[year.Index];
                    RegulatoryClass    vehRC    = vehs[j].RegClass;
                    if (vehs[j].HEVType == HEVType.PureElectric)
                    {
                        vmd.RelativeLossOfValue = vehMsrp * econData.RelativeValueLoss;
                    }
                    vmd.TaxesAndFees  = vehMsrp * econData.TaxesAndFees;
                    vmd.FinancingCost = vehMsrp * econData.Financing;
                    vmd.InsuranceCost = vehMsrp * econData.Insurance;
                    mmd.RelativeLossOfValue[vehRC] += (vehSales * vmd.RelativeLossOfValue);
                    mmd.TaxesAndFees       [vehRC] += (vehSales * vmd.TaxesAndFees       );
                    mmd.FinancingCost      [vehRC] += (vehSales * vmd.FinancingCost      );
                    mmd.InsuranceCost      [vehRC] += (vehSales * vmd.InsuranceCost      );
                }
                mmd.LossOfValue *= (1 - econData.DiscountRates[0]);
                mmd.TotalConsumerCosts = mmd.TechCost + mmd.Fines + mmd.TaxesAndFees + mmd.FinancingCost + mmd.InsuranceCost +
                                         mmd.MaintenanceCost + mmd.RepairCost + mmd.LossOfValue + mmd.RelativeLossOfValue;
                mmd.TotalSocialCosts   = mmd.TechCost +
                                         mmd.MaintenanceCost + mmd.RepairCost + mmd.LossOfValue + mmd.RelativeLossOfValue;
            }
            this._effectsModel.Begin(scen, year, bData, data, this.Settings);
            if (this.Settings.OperatingModes.MultiYearModeling)
            {   
                if (year.Year > this.MinYear)
                {   
                    this.PostProcessModelYear(scen, this.ModelYears[year.Year - this.MinYear - 1]);
                }
            }
            if (year.Year == this._year.Year)
            {
                this.OnModelYearCompleted();
            }
        }
        #endregion
        #region 
        protected virtual void ManufacturerLoop(Scenario scen, ModelYear year)
        {
            for (int i = 0; i < this.MfrCount; i++)
            {   
                this.RunManufacturer(scen, year, i);
                if (this.AbortRequested) { break; }
            } 
        }
        protected virtual void RunManufacturer(Scenario scen, ModelYear year, int mfrIndex)
        {
            this.PreProcessManufacturer(scen, year, mfrIndex);
            int                       yrIndex      = year.Index;
            Industry                  data         = this.GetData(scen, year);
            Manufacturer              mfr          = data.Manufacturers[mfrIndex];
            Manufacturer.CModelData   mmd          = mfr.ModelData;
            Manufacturer.CDescription mfrDescr     = mfr.Description;
            ModelingApproach          approach     = this.Settings.OperatingModes.ModelingApproach;
            bool                      noFines      = this.Settings.OperatingModes.NoFines || !mfrDescr.WillingToPayFines[year.Index];
            double                    mfrTechCost  = mmd.TechCost.TotalAll;
            double                    mfrFines     = mmd.Fines   .TotalAll;
            bool                      ignoredReset = false;
            bool                      overcomply   = (this.Settings.OperatingModes.Overcomply && mfrFines <= 0);
            bool                      overcomplyAllowInefficient = !this.Settings.OperatingModes.OvercomplyCostEffectiveOnly;
            int                       overcomplyLookAhead        =  this.Settings.OperatingModes.OvercomplyLookAheadYears;
            ComplianceFinding finding = null;
            mmd.CFExhausted = false;
            while (mfrFines > 0 || overcomply)
            {   
                finding = this.ComplianceFindingLoop(scen, year, mfrIndex, noFines, ref approach, overcomply);
                if (finding != null && finding.IsValid)
                {   
                    bool applyFinding = (mfrFines > 0);
                    if (!applyFinding && overcomply)
                    {   
                        applyFinding = (overcomplyAllowInefficient || finding.IsEfficient);
                        if (applyFinding)
                        {   
                            if (overcomplyLookAhead > 0)
                            {
                                ModelYear overcomplyYear = new ModelYear(Math.Min(year.Year + overcomplyLookAhead, this.MaxYear));
                                RCDouble overcomplyNewFines = Standards.TestFines(mfr, scen, overcomplyYear, this.Settings.Parameters,
                                    finding._newCafe, finding._techImpacts, finding._vehCount);
                                if (overcomplyNewFines.TotalAll <= 0) { applyFinding = false; }
                            }
                        }
                    } 
                    if (applyFinding)
                    {
                        finding.ApplyFinding();
                        this.Recomply(scen, year, mfr);
                        CreditTrading.ExamineCredits(this.GetData(scen), year.Index, mfrIndex, this.Settings);
                        this.Recomply(scen, year, mfr);
                    }
                    else { break; }
                } 
                else
                {
                    if (finding == null)
                    {   
                        if (ignoredReset)
                        {   
                            mmd.CFExhausted = true;
                            break;
                        }
                        else
                        {   
                            List<Vehicle> mfrVehs = mfr.Vehicles;
                            for (int i = 0, vehCount = mfrVehs.Count; i < vehCount; i++)
                            {
                                Vehicle.CModelData vmd = mfrVehs[i].ModelData;
                                for (int j = 0; j < TI.TechnologyCount; j++) { vmd.TechIgnored[j] = false; }
                            }
                            ignoredReset = true;    
                        }
                    }
                    else 
                    {   
                        throw new Exception("The compliance finding is not valid!");
                    }
                }
                mfrTechCost = mmd.TechCost.TotalAll;
                mfrFines    = mmd.Fines   .TotalAll;
                if (mfrFines <= 0)
                {   
                    overcomply = this.Settings.OperatingModes.Overcomply;
                }
                if (this.AbortRequested) { break; }
            } 
            this.PostProcessManufacturer(scen, year, mfrIndex);
        }
        protected virtual void PreProcessManufacturer(Scenario scen, ModelYear year, int mfrIndex)
        {
            Manufacturer mfr = this.GetData(scen, year).Manufacturers[mfrIndex];
            this._mfr = mfr;
            this.OnManufacturerStarted();
            this.Recomply(scen, year, mfr);
            CreditTrading.ExamineCredits(this.GetData(scen), year.Index, mfrIndex, this.Settings);
            this.Recomply(scen, year, mfr);
        }
        protected virtual void PostProcessManufacturer(Scenario scen, ModelYear year, int mfrIndex)
        {
            if (this.AbortRequested) { return; }
            int          yrIndex = year.Index;
            Industry     data    = this.GetData(scen, year);
            Manufacturer mfr     = data.Manufacturers[mfrIndex];
            this.Recomply(scen, year, mfr);
            if (this.Settings.OperatingModes.MultiYearModeling && this.Settings.OperatingModes.MultiYearStartYear <= year.Year)
            {   
                List<Vehicle> mfrVehs = mfr.Vehicles;
                for (int i = 0, mfrVehCount = mfrVehs.Count; i < mfrVehCount; i++)
                {   
                    Vehicle              veh = mfrVehs[i];
                    Vehicle.CDescription vd  = veh.Description;
                    Vehicle.CModelData   vmd = veh.ModelData;
                    bool atRedesign, inShadow;
                    veh.GetRedesignState(year, out atRedesign, out inShadow);
                    if (atRedesign)
                    {
                        for (int j = 0; j < TI.TechnologyCount; j++)
                        {   
                            if (TI.IsTiedToRedesign(j) || TI.IsTiedToRefresh(j))
                            {
                                if (vmd.TechEnabled[j] && !vmd.TechUsed[j])
                                {   
                                    vmd.TechLingerYear = year;
                                    vmd.TechLingering[j] = true;
                                }
                            }
                        } 
                    } 
                }
                for (int i = this.MinYearIndex; i < yrIndex; i++)
                {
                    ModelYear prevYear = this.ModelYears[i - this.MinYearIndex];
                    Industry prevData = this.GetData(scen, prevYear);
                    Manufacturer lMfr = prevData.Manufacturers[mfrIndex];
                    this.Recomply(scen, prevYear, lMfr);
                }
            } 
            this.OnManufacturerCompleted();
        }
        protected virtual void Recomply(Scenario scen, ModelYear year, Manufacturer mfr)
        {
            Manufacturer.CModelData mmd = mfr.ModelData;
            Standards.CalcStandard(mfr, scen, year);
            Standards.CalcCAFE    (mfr, scen, year, this.Settings.Parameters);      
            Standards.CalcCredits (mfr, scen, year);
            Standards.CalcFines   (mfr,             this.Settings.Parameters);
        }
        #endregion
        #region 
        protected virtual ComplianceFinding ComplianceFindingLoop(Scenario scen, ModelYear year, int mfrIndex, bool noFines,
            ref ModelingApproach approach, bool overcomply)
        {
            int                     snIndex = scen.Index;
            int                     yrIndex = year.Index;
            Industry                data    = this.GetData(scen, year);
            Manufacturer            mfr     = data.Manufacturers[mfrIndex];
            Manufacturer.CModelData mmd     = mfr.ModelData;
            ComplianceFinding finding = null;
            bool lowCostFirst    = (approach == ModelingApproach.LowCostFirst);
            bool allowSecondBest = !lowCostFirst;   
            bool loop            = true;
            while (loop)
            {   
                finding = this._lcf.GetNextFinding(year, this.GetData(scen), mfrIndex, lowCostFirst, allowSecondBest, overcomply);
                if (lowCostFirst && (finding == null || !finding.IsEfficient))
                {   
                    if (this.Settings.OperatingModes.AutoToggleApproach)
                    {   
                        approach        = ModelingApproach.EconomicEfficiency;
                        lowCostFirst    = false;
                        allowSecondBest = true;
                        continue;
                    }
                    else if (!allowSecondBest)
                    {   
                        allowSecondBest = true;
                        continue;
                    }
                } 
                double mfrTechCost = mmd.TechCost.TotalAll;
                double mfrFines    = mmd.Fines   .TotalAll;
                if (finding == null || !finding.IsValid || noFines || finding.IsEfficient)
                {   
                    loop = false;
                }
                else
                {   
                    finding = null;
                    loop    = false;
                }
            } 
            return finding;
        }
        #endregion
        #endregion
        #endregion
        #region 
        #region 
        public override IModelingProgress Progress
        {
            get
            {
                bool   xlReporting          = (this.XlReportGenerator  != null && this.XlReportGenerator .Reporting);
                bool   csvReporting         = (this.CsvReportGenerator != null && this.CsvReportGenerator.Reporting);
                bool   reporting            =  xlReporting || csvReporting;
                string xlReportingProgress  = (xlReporting) ? this.XlReportGenerator.ReportingProgress : null;
                string csvReportingProgress = (csvReporting) ? this.CsvReportGenerator.ReportingProgress : null;
                string reportingProgress    =  xlReportingProgress + ((xlReporting) ? "\n" : "") + csvReportingProgress;
                return (this.Running || reporting) ? new ModelingProgress(this._scen, this._year, this._mfr, reportingProgress) : null;
            }
        }
        #endregion
        #endregion
        #region 
        [NonSerialized] internal LinearComplianceFinder _lcf;
        [NonSerialized] Effects _effectsModel;
        #endregion
    }
}

