#region << Using Directives >>
using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using Volpe.Cafe.Data;
using Volpe.Cafe.IO;
using Volpe.Cafe.Settings;
using Volpe.Cafe.Utils;
using TI = Volpe.Cafe.TechnologyIndexes;
using TA = Volpe.Cafe.Model.TechnologyApplicability;
#endregion
namespace Volpe.Cafe.Model
{
    [Serializable]
    public class ComplianceFinding
    {
        #region 
        public ComplianceFinding()
        {
            this._capacity                 = 8;
            this._vehs                     = new Vehicle   [this._capacity];
            this._ignore                   = new int       [this._capacity];
            this._techKeys                 = new ulong     [this._capacity];
            this._techLingerKeys           = new ulong     [this._capacity];
            this._lingerYear               = new int       [this._capacity];
            this._techImpacts              = new TechImpact[this._capacity];
            this._yearOfInitialApplication = new int       [this._capacity][];
            this._committedERPCosts        = new bool      [TI.TechnologyCount];
            this.Reset();
        }
        #endregion
        #region 
        #region 
        public override string ToString()
        {
            bool valid = this.IsValid;
            if (valid)
            {
                Manufacturer.CModelData mmd = this._vehs[0].Manufacturer.ModelData;
                return "ComplianceFinding # " + this._index +
                    ": {\nIsValid: true\nIsEfficient: "  + this.IsEfficient +
                    "\nEfficiency: "    + this._efficiency         +
                    "\nTechCost: "      + this._techCost           +
                    "\nRPECost: "       + this._totalDeltaRPE      +
                    "\nDeltaFines: "    + this._totalDeltaFines    +
                    "\nFuelSavings: "   + this._totalFuelSavings   +
                    "\nAffectedSales: " + this._totalAffectedSales +
                    "\nCAFE (old, new): {"  +
                    "\n\t  Domestic Auto: " + this._curCafe.DomesticAuto  + ", " + this._newCafe.DomesticAuto  +
                    "\n\t  Imported Auto: " + this._curCafe.ImportedAuto  + ", " + this._newCafe.ImportedAuto  +
                    "\n\t  Light Truck: "   + this._curCafe.LightTruck    + ", " + this._newCafe.LightTruck    +
                    "\n\t  Unregulated: "   + this._curCafe.Unregulated   + ", " + this._newCafe.Unregulated   +
                    "\n\t}\nFines (old, new): {" +
                    "\n\t  Domestic Auto: " + this._curFines.DomesticAuto + ", " + this._newFines.DomesticAuto +
                    "\n\t  Imported Auto: " + this._curFines.ImportedAuto + ", " + this._newFines.ImportedAuto +
                    "\n\t  Light Truck: "   + this._curFines.LightTruck   + ", " + this._newFines.LightTruck   +
                    "\n\t  Unregulated: "   + this._curFines.Unregulated  + ", " + this._newFines.Unregulated  +
                    "\n\t}\n}";
            }
            else { return "ComplianceFinding # " + this._index + ": {\nValid: false}"; }
        }
        #endregion
        #region 
        public void Initialize(Scenario scen, ModelYear year, int startYear, int endYear, Industry[] modelYearData, int mfrIndex,
            ModelingSettings settings, LogWriter logWriter, TechnologyType techType, List<Technology> techGroup)
        {
            this._scen          = scen;
            this._year          = year;
            this._yrIndex       = year.Index;
            this._yrYear        = year.Year;
            this._startYear     = startYear;
            this._endYear       = endYear;
            this._modelYearData = modelYearData;
            this._mfrIndex      = mfrIndex;
            this._settings      = settings;
            this._parameters    = settings.Parameters;
            this._multiYear     = settings.OperatingModes.MultiYearModeling;
            this._ignorePhaseIn = settings.OperatingModes.IgnorePhaseIn;
            this._backfill      = settings.OperatingModes.Backfill;
            this._splitShared   = settings.OperatingModes.SplitSharedEngsOrTrns;
            this._splitMixed    = settings.OperatingModes.SplitConflictingEngsOrTrns;
            this._skipShared    = settings.OperatingModes.SkipSharedEngsOrTrns;
            this._skipMixed     = settings.OperatingModes.SkipConflictingEngsOrTrns;
            this._logWriter     = logWriter;
            this._isEngOrTrn    = (techType == TechnologyType.Engine || techType == TechnologyType.Transmission);
            this._techType      = techType;
            this._techGroup     = techGroup;
            this._techCount     = techGroup.Count;
            this.Reset();
        }
        public void Reset()
        {
            this._index              = -1;
            this._vehCount           = 0;
            this._techIndex          = 0;
            this._hasVehs            = false;
            this._isRcMix            = false;
            this._isEnablesMix       = false;
            this._isSplitCf          = false;
            this._isApplied          = false;
            this._efficiency         = double.NaN;
            this._techCost           = double.NaN;
            this._newCafe            = RCDouble.Zero;
            this._newFines           = RCDouble.Zero;
            this._effectiveNewFines  = RCDouble.Zero;
            this._curCafe            = RCDouble.Zero;
            this._curFines           = RCDouble.Zero;
            this._totalDeltaRPE      = 0;
            this._totalDeltaFines    = 0;
            this._totalFuelSavings   = 0;
            this._totalAffectedSales = 0;
        }
        public void Clear()
        {
            this._scen          = null;
            this._year          = null;
            this._yrIndex       = 0;
            this._yrYear        = 0;
            this._startYear     = 0;
            this._endYear       = 0;
            this._modelYearData = null;
            this._mfrIndex      = 0;
            this._settings      = null;
            this._parameters    = null;
            this._overcomply    = false;
            this._multiYear     = false;
            this._ignorePhaseIn = false;
            this._backfill      = false;
            this._splitShared   = false;
            this._splitMixed    = false;
            this._skipShared    = false;
            this._skipMixed     = false;
            this._logWriter     = LogWriter.Empty;
            this._isEngOrTrn    = false;
            this._techType      = TechnologyType.Other;
            this._techGroup     = null;
            this._techCount     = 0;
            this.Reset();
        }
        public void EnsureCapacity(int capacity)
        {
            if (this._capacity < capacity)
            {
                this._capacity = ((capacity / 8) + 1) * 8;
                this.EnsureCapacity();
            }
        }
        void EnsureCapacity()
        {
            if (this._techKeys.Length < this._capacity)
            {
                this._vehs                     = new Vehicle   [this._capacity];
                this._ignore                   = new int       [this._capacity];
                this._techKeys                 = new ulong     [this._capacity];
                this._techLingerKeys           = new ulong     [this._capacity];
                this._lingerYear               = new int       [this._capacity];
                this._techImpacts              = new TechImpact[this._capacity];
                this._yearOfInitialApplication = new int       [this._capacity][];
            }
            int impactLen = this._endYear - ModelYear.MinYear + 1;
            for (int i = 0; i < this._capacity; i++)
            {
                if (this._techImpacts[i] == null)
                {
                    this._techImpacts[i] = new TechImpact(impactLen);
                }
                if (this._yearOfInitialApplication[i] == null)
                {
                    this._yearOfInitialApplication[i] = new int[TI.TechnologyCount];
                }
            }
        }
        public void CopyFrom(ComplianceFinding cf)
        {
            this._scen          = cf._scen;
            this._year          = cf._year;
            this._yrIndex       = cf._yrIndex;
            this._yrYear        = cf._yrYear;
            this._startYear     = cf._startYear;
            this._endYear       = cf._endYear;
            this._modelYearData = cf._modelYearData;    
            this._mfrIndex      = cf._mfrIndex;
            this._settings      = cf._settings;
            this._parameters    = cf._parameters;
            this._overcomply    = cf._overcomply;
            this._multiYear     = cf._multiYear;
            this._ignorePhaseIn = cf._ignorePhaseIn;
            this._backfill      = cf._backfill;
            this._splitShared   = cf._splitShared;
            this._splitMixed    = cf._splitMixed;
            this._skipShared    = cf._skipShared;
            this._skipMixed     = cf._skipMixed;
            this._logWriter     = cf._logWriter;
            this._isEngOrTrn    = cf._isEngOrTrn;
            this._techType      = cf._techType;
            this._techGroup     = cf._techGroup;
            this._techCount     = cf._techCount;
            this._index              = cf._index;
            this._vehCount           = cf._vehCount;
            this._capacity           = cf._capacity;
            this._techIndex          = cf._techIndex;
            this._hasVehs            = cf._hasVehs;
            this._isRcMix            = cf._isRcMix;
            this._isEnablesMix       = cf._isEnablesMix;
            this._isSplitCf          = cf._isSplitCf;
            this._isApplied          = cf._isApplied;
            this._efficiency         = cf._efficiency;
            this._techCost           = cf._techCost;
            this._newCafe            = cf._newCafe;
            this._newFines           = cf._newFines;
            this._effectiveNewFines  = cf._effectiveNewFines;
            this._curCafe            = cf._curCafe;
            this._curFines           = cf._curFines;
            this._totalDeltaRPE      = cf._totalDeltaRPE;
            this._totalDeltaFines    = cf._totalDeltaFines;
            this._totalFuelSavings   = cf._totalFuelSavings;
            this._totalAffectedSales = cf._totalAffectedSales;
            this.EnsureCapacity();
            Array.Copy(cf._vehs          , 0, this._vehs          , 0, this._vehCount);
            Array.Copy(cf._ignore        , 0, this._ignore        , 0, this._vehCount);
            Array.Copy(cf._techKeys      , 0, this._techKeys      , 0, this._vehCount);
            Array.Copy(cf._techLingerKeys, 0, this._techLingerKeys, 0, this._vehCount);
            Array.Copy(cf._lingerYear    , 0, this._lingerYear    , 0, this._vehCount);
            for (int i = 0; i < this._vehCount; i++)
            {
                cf._techImpacts[i].CopyTo(this._techImpacts[i]);
                Array.Copy(cf._yearOfInitialApplication[i], 0, this._yearOfInitialApplication[i], 0, TI.TechnologyCount);
            }
        }
        public void ExamineFinding(int techIdx, List<Vehicle> vehs, int vehIdx, ref int cfCount, RegulatoryClass regClass, bool overcomply)
        {
            this.Reset();
            this.ParseVehTechsLists(techIdx, vehs, vehIdx, regClass);
            this._hasVehs = false;
            for (int i = 0; i < this._vehCount; i++)
            {
                if (this._vehs[i].RegClass != RegulatoryClass.Unregulated) { this._hasVehs = true; break; }
            }
            if (!this._hasVehs) { return; }     
            bool skipShared       = this._isRcMix && this._skipShared;
            bool skipMixedEnables = this._isEnablesMix && this._skipMixed;
            if (skipShared || skipMixedEnables) { return; }
            bool splitShared       = this._isRcMix      && this._splitShared && this._isEngOrTrn;
            bool splitMixedEnables = this._isEnablesMix && this._splitMixed  && this._isEngOrTrn;
            this._overcomply = overcomply;
            if (splitShared)
            {   
                this.SplitShared(splitMixedEnables, ref cfCount);
            }
            else if (splitMixedEnables)
            {   
                this.SplitMixedEnables(true, ref cfCount);
            }
            else
            {   
                this.CalcEfficiency();
                this._index = cfCount++;
            }
        }
        void ParseVehTechsLists(int techIdx, List<Vehicle> vehs, int vehIdx, RegulatoryClass regClass)
        {
            int allVehCount = (vehIdx == -1) ? vehs.Count : 1;
            this.EnsureCapacity(allVehCount);
            this._techIndex = techIdx;
            Manufacturer            mfr = this._modelYearData[this._yrIndex].Manufacturers[this._mfrIndex];
            Manufacturer.CModelData mmd = mfr.ModelData;
            int yrLen = this._endYear - ModelYear.MinYear + 1;
            int vIdxOff = (vehIdx == -1) ? 0 : vehIdx;
            for (int i = 0; i < allVehCount; i++)
            {
                Vehicle veh = vehs[i + vIdxOff];
                int lYr = (veh.ModelData.TechLingerYear == null) ? -1 : veh.ModelData.TechLingerYear.Index;
                Manufacturer.CModelData lMmd = (lYr == -1) ? null : this._modelYearData[lYr].Manufacturers[this._mfrIndex].ModelData;
                bool lingering;
                if ((!this._splitShared || regClass == (RegulatoryClass)(-1) || veh.RegClass == regClass) &&
                    TA.IsTechValid(mmd, veh.ModelData, veh.RegClass, this._techGroup[techIdx].Index, this._ignorePhaseIn,
                    false, lMmd, out lingering))
                {   
                    this._vehs       [this._vehCount] = veh;
                    this._ignore     [this._vehCount] = 0;
                    this._techKeys   [this._vehCount] = TA.TechKeys[techIdx];
                    if (this._techImpacts[this._vehCount] == null)
                    {
                        this._techImpacts[this._vehCount] = new TechImpact(yrLen);
                    }
                    if (this._yearOfInitialApplication[this._vehCount] == null)
                    {
                        this._yearOfInitialApplication[this._vehCount] = new int[TI.TechnologyCount];
                    }
                    if (this._multiYear && lingering)
                    {   
                        this._techLingerKeys[this._vehCount] = TA.TechKeys[techIdx];
                    }
                    else
                    {   
                        this._techLingerKeys[this._vehCount] = 0;
                    }
                    if (this._backfill)
                    {
                        this.ParseVehTechsLists_Backfill(techIdx, veh, mmd);
                    }
                    this._vehCount++;
                }
            } 
            if (this._vehCount > 0)
            {   
                RegulatoryClass veh0rc = this._vehs[0].RegClass;
                for (int i = 1; i < this._vehCount; i++)
                {   
                    if (this._vehs[i].RegClass != veh0rc) { this._isRcMix = true; break; }
                }
                if (this._vehCount != allVehCount) { this._isEnablesMix = true; }
                else
                {   
                    for (int i = 1; i < this._vehCount; i++)
                    {   
                        if (this._techKeys[i] != this._techKeys[0]) { this._isEnablesMix = true; break; }
                    }
                }
            } 
        }
        void ParseVehTechsLists_Backfill(int techIdx, Vehicle veh, Manufacturer.CModelData mmd)
        {
            TechnologyClass    techClass = veh.TechnologyClass;
            RegulatoryClass    regClass  = veh.RegClass;
            Vehicle.CModelData vmd       = veh.ModelData;
            for (int i = techIdx - 1; i >= 0; i--)
            {   
                int i2 = -1, iTmp = i;  
                switch (this._techGroup[i].Index)
                {   
                    case TI.DVVLD: i2 = this.FindTechInGroup(TI.CVVL); break;
                    case TI.CVVL:
                        continue;       
                }
                if (i2 != -1)
                {   
                    Industry yrData = this._modelYearData[this._yrIndex];
                    if (TA.GetCost(this._techGroup[i2], veh, yrData, this._settings, this._year, this._year.Year, this._startYear, this._endYear, this._techGroup, this._techKeys[this._vehCount]) <
                        TA.GetCost(this._techGroup[i ], veh, yrData, this._settings, this._year, this._year.Year, this._startYear, this._endYear, this._techGroup, this._techKeys[this._vehCount]))
                    {   
                        Interaction.Swap(ref i, ref i2);
                    }
                }
                for (int j = 0; j < 2; j++)
                {   
                    int bftIndex = this._techGroup[i].Index;    
                    int lYr = (veh.ModelData.TechLingerYear == null) ? -1 : veh.ModelData.TechLingerYear.Index;
                    Manufacturer.CModelData lMmd = (lYr == -1) ? null : this._modelYearData[lYr].Manufacturers[this._mfrIndex].ModelData;
                    bool lingering;
                    bool isTechValid = TA.IsTechValid(mmd, vmd, regClass, bftIndex, this._ignorePhaseIn, true, lMmd, out lingering);
                    bool canBackfill = TA.CanBackfill(i, this._techGroup, this._techKeys[this._vehCount], vmd.TechUsed);
                    if (isTechValid && canBackfill)
                    {
                        this._techKeys[this._vehCount] += TA.TechKeys[i];
                        if (this._multiYear && lingering)
                        {
                            this._techLingerKeys[this._vehCount] += TA.TechKeys[i];
                        }
                    }
                    if (i2 == -1) { break; }    
                    i = i2;                     
                }
                i = iTmp;
            } 
        }
        void SplitShared(bool splitMixedEnables, ref int cfCount)
        {
            double eff   = double.NaN;
            int    effRc = 0;
            for (int i = 1; i <= 3; i++)
            {   
                RegulatoryClass regClass = (RegulatoryClass)i;
                int firstRcVeh = -1;    
                int rcVehCount =  0;    
                for (int j = 0; j < this._vehCount; j++)
                {
                    if (this._vehs[j].RegClass == regClass)
                    {   
                        this._ignore[j] = 0;                        
                        if (firstRcVeh == -1) { firstRcVeh = j; }   
                        rcVehCount++;                               
                    }
                    else { this._ignore[j] = 1; }
                }
                if (rcVehCount == 0) { continue; }
                bool isEnablesMix = false;
                if (splitMixedEnables)
                {
                    for (int j = firstRcVeh + 1; j < this._vehCount; j++)
                    {   
                        if (this._ignore[j] == 0 && this._techKeys[j] != this._techKeys[firstRcVeh])
                        {   
                            isEnablesMix = true; break;
                        }
                    }
                }
                this._isEnablesMix = isEnablesMix;
                if (isEnablesMix) { this.SplitMixedEnables(i == 1, ref cfCount); }
                else              { this.CalcEfficiency(); }
                if (double.IsNaN(eff) || this._efficiency < eff)
                {
                    eff   = this._efficiency;
                    effRc = i;
                }
            } 
            if (effRc > 0)
            {
                if (effRc < 3)
                {   
                    RegulatoryClass regClass = (RegulatoryClass)effRc;
                    for (int i = 0; i < this._vehCount; i++)
                    {
                        this._ignore[i] = (this._vehs[i].RegClass == regClass) ? 0 : 1;
                    }
                    this.CalcEfficiency();
                }
                this._index     = cfCount++;
                this._isSplitCf = true;
            }
        }
        void SplitMixedEnables(bool sortLists, ref int cfCount)
        {
            if (sortLists)
            {   
                for (int i = this._vehCount - 1; i >= 0; i--)
                {
                    for (int j = 1; j <= i; j++)
                    {
                        if (this._techKeys[j - 1] > this._techKeys[j])
                        {   
                            this.Swap(this._vehs          , j - 1, j);
                            this.Swap(this._ignore        , j - 1, j);
                            this.Swap(this._techKeys      , j - 1, j);
                            this.Swap(this._techLingerKeys, j - 1, j);
                        }
                    } 
                } 
            }
            double eff    = double.NaN;     
            ulong  effKey = 0;              
            ulong  key    = 0;              
            for (int i = 0; i < this._vehCount; i++)
            {
                if ((this._ignore[i] & 1) != 1)
                {   
                    if (key == this._techKeys[i]) { this._ignore[i] -= 2; }
                    else
                    {   
                        if (key != 0)
                        {   
                            this.CalcEfficiency();
                            if (double.IsNaN(eff) || this._efficiency < eff)
                            {   
                                eff    = this._efficiency;
                                effKey = key;
                            }
                        }
                        key = this._techKeys[i];
                        for (int j = 0; j < this._vehCount; j++) { this._ignore[j] |= 2; }
                        this._ignore[i] -= 2;   
                    }
                } 
            } 
            this.CalcEfficiency();
            if (double.IsNaN(eff) || this._efficiency < eff)
            {   
                eff    = this._efficiency;
                effKey = key;
            }
            if (effKey != 0)
            {
                for (int i = 0; i < this._vehCount; i++)
                {   
                    if (this._techKeys[i] == effKey) { this._ignore[i] &= 1; }
                    else                             { this._ignore[i] |= 2; }
                }
                this.CalcEfficiency();
                this._index     = cfCount++;
                this._isSplitCf = true;
            }
        }
        int FindTechInGroup(int techIdx)
        {
            for (int i = 0; i < this._techCount; i++)
            {
                if (this._techGroup[i].Index == techIdx) { return i; }
            }
            return -1;
        }
        void Swap(Vehicle[] arr, int i, int j) { Vehicle tmp = arr[i]; arr[i] = arr[j]; arr[j] = tmp; }
        void Swap(ulong  [] arr, int i, int j) { ulong   tmp = arr[i]; arr[i] = arr[j]; arr[j] = tmp; }
        void Swap(int    [] arr, int i, int j) { int     tmp = arr[i]; arr[i] = arr[j]; arr[j] = tmp; }
        #endregion 
        #region 
        void CalcEfficiency()
        {
            this._efficiency = double.NaN;
            bool hasVehs = false;   
            for (int i = 0; i < this._vehCount; i++)
            {
                if (this._ignore[i] == 0 && this._vehs[i].RegClass != RegulatoryClass.Unregulated)
                {   
                    hasVehs = true; break;
                }
            }
            if (!hasVehs) { return; }   
            if (this._multiYear)
            {
                for (int i = 0; i < this._vehCount; i++)
                {
                    if (this._techLingerKeys[i] != 0)
                    {
                        this._lingerYear[i] = this._vehs[i].ModelData.TechLingerYear.Year;
                    }
                }
            }
            Industry     data = this._modelYearData[this._yrIndex];
            Manufacturer mfr  = data.Manufacturers[this._mfrIndex];
            bool mfrInCompliance = true;
            for (int i = 0; i < this._vehCount; i++)
            {
                if (this._ignore[i] == 0)
                {   
                    RegulatoryClass regClass = this._vehs[i].RegClass;
                    if (mfr.ModelData.Fines[regClass] > 0) { mfrInCompliance = false; break; }
                }
            }
            if (mfrInCompliance && !this._overcomply) { return; }
            this.CalcTechImpacts(data.ModelData);
            this._newCafe  = Standards.TestCAFE (mfr, this._scen, this._year, this._parameters,                this._techImpacts, this._vehCount, false);
            this._newFines = Standards.TestFines(mfr, this._scen, this._year, this._parameters, this._newCafe, this._techImpacts, this._vehCount);
            RCDouble effectiveNewCafe = Standards.TestCAFE (mfr, this._scen, this._year, this._parameters,                                    this._techImpacts, this._vehCount, true);
            this._effectiveNewFines   = Standards.TestFines(mfr, this._scen, this._year, this._parameters, effectiveNewCafe.CreateRounded(1), this._techImpacts, this._vehCount);
            this._curCafe  = Standards.TestCAFE (mfr, this._scen, this._year, this._parameters,                null, 0, false);
            this._curFines = Standards.TestFines(mfr, this._scen, this._year, this._parameters, this._curCafe, null, 0);
            RCDouble deltaFines = this._curFines - this._newFines;
            double totalDeltaFines = deltaFines.TotalAll;
            if (totalDeltaFines <= 0 && !this._overcomply) { this.IgnoreTechnologies(); return; }
            Manufacturer.CModelData mmd = mfr.ModelData;
            KMYType kmy = (mfrInCompliance) ? mmd.KMY_OC[this._yrIndex] : mmd.KMY[this._yrIndex];
            double techCost      = 0;
            double affectedSales = 0;
            double fcSavings     = 0;
            double deltaRpe      = 0;
            double lossOfValue   = 0;
            for (int i = 0; i < this._vehCount; i++)
            {
                TechImpact impact = this._techImpacts[i];
                if (!impact.Ignore)
                {
                    for (int yrIndex = 0; yrIndex <= this._yrIndex; yrIndex++)
                    {
                        if (impact.HasImpact[yrIndex])
                        {
                            Vehicle         veh      = impact.Vehicle[yrIndex];
                            RegulatoryClass regClass = veh.RegClass;
                            FuelValue       vehFE    = veh.Description.FuelEconomy;
                            FuelValue       vehShare = veh.Description.FuelShare;
                            FuelValue       newFE    = impact.NewFuelEconomy[yrIndex];
                            FuelValue       newShare = impact.NewFuelShare  [yrIndex];
                            double          sales    = veh.Description.Sales[yrIndex];
                            FuelValue gap = this._parameters.EconomicData.TestVsOnRoadGap;
                            double vehFuelCost = Standards.GetVehicleFuelCost(vehFE, vehShare, gap, kmy, this._scen.MultiFuelCAFE[yrIndex]);
                            double newFuelCost = Standards.GetVehicleFuelCost(newFE, newShare, gap, kmy, this._scen.MultiFuelCAFE[yrIndex]);
                            techCost      += sales * impact.Cost[yrIndex];
                            affectedSales += sales;
                            fcSavings     += sales * (vehFuelCost - newFuelCost);
                            lossOfValue   += sales * impact.LossOfValue[yrIndex];
                        } 
                    } 
                } 
            } 
            if (fcSavings <= 0 && totalDeltaFines <= 0) { this.IgnoreTechnologies(); return; }
            deltaRpe = techCost * 0.77;
            double numerator   = deltaRpe - totalDeltaFines - fcSavings + lossOfValue;
            double denominator = affectedSales;
            this._efficiency   = numerator / denominator;
            this._techCost     = techCost;
            this._totalDeltaRPE      = deltaRpe;
            this._totalDeltaFines    = totalDeltaFines;
            this._totalFuelSavings   = fcSavings;
            this._totalAffectedSales = affectedSales;
        }
        Vehicle CalcTechImpacts_GetVehicle(int vehIndex, int yrIndex)
        {
            Vehicle veh = this._vehs[vehIndex];
            if (yrIndex == this._yrIndex)
            {   
                return veh;
            }
            else
            {   
                Manufacturer lMfr = this._modelYearData[yrIndex].Manufacturers[this._mfrIndex];
                List<Vehicle> lVehs = lMfr.Vehicles;
                int     lCode = veh.Description.Code;
                Vehicle lVeh = null;
                for (int i = 0, count = lVehs.Count; i < count; i++)
                {
                    if (lVehs[i].Description.Code == lCode) { lVeh = lVehs[i]; break;  }
                }
                if (lVeh == null)
                {   
                    return null;
                }
                if (lCode == veh.Description.Code)
                {   
                    return lVeh;
                }
                if ((TA.HasEngine      (this._techGroup) && lVeh.Description.EngineCode       != veh.Description.EngineCode) ||
                    (TA.HasTransmission(this._techGroup) && lVeh.Description.TransmissionCode != veh.Description.TransmissionCode))
                {   
                    return null;
                }
                return lVeh;
            } 
        }
        void CalcTechImpacts(Industry.CModelData imd)
        {
            for (int i = 0; i < this._vehCount; i++)
            {   
                this._techImpacts[i].Reset();
                if (this._ignore[i] != 0) { this._techImpacts[i].Ignore = true; continue; }
                Vehicle curVeh        = this._vehs          [i];                            
                ulong   curTechKey    = this._techKeys      [i];                            
                ulong   lingerTechKey = this._techLingerKeys[i];                            
                bool    lingering     = (lingerTechKey != 0);                               
                int     lingerYrYear  = (lingering) ? this._lingerYear[i] : this._yrYear;   
                int     lingerYrIndex = lingerYrYear - ModelYear.MinYear;                   
                TechImpact impact = this._techImpacts[i];
                Array.Clear(this._yearOfInitialApplication[i], 0, TI.TechnologyCount);
                for (int j = lingerYrIndex; j <= this._yrIndex; j++)
                {   
                    Vehicle veh = this.CalcTechImpacts_GetVehicle(i, j);
                    if (veh == null) { impact.Ignore = true; break; }
                    ModelYear            year      = ModelYear.NewModelYearFromIndex(j);    
                    Industry             data      = this._modelYearData[j];
                    Vehicle.CDescription vd        = veh.Description;
                    Vehicle.CModelData   vmd       = veh.ModelData;
                    TechnologyClass      techClass = curVeh.TechnologyClass;
                    RegulatoryClass      regClass  = curVeh.RegClass;
                    ulong                techKey   = (j == this._yrIndex) ? curTechKey : lingerTechKey;
                    bool shev1Used = vmd.TechUsed[TI.SHEV1];
                    if (techKey != 0)
                    {   
                        impact.HasImpact[j] = true;
                        impact.Vehicle  [j] = veh;      
                        impact.NewFuelShare   [j] = veh.Description.FuelShare;
                        impact.NewFuelEconomy [j] = veh.Description.FuelEconomy;
                        impact.NewFuelEconomy2[j] = veh.Description.FuelEconomy2;
                        Array.Clear(this._committedERPCosts, 0, TI.TechnologyCount);
                        for (int k = 0; k < this._techCount; k++)
                        {
                            if ((techKey & TA.TechKeys[k]) != 0)
                            {   
                                Technology tech = this._techGroup[k];
                                impact.LossOfValue[j] += tech.Attributes[veh.TechnologyClass].LossOfValue;
                                if (this._yearOfInitialApplication[i][tech.Index] == 0) { this._yearOfInitialApplication[i][tech.Index] = j; }
                                double erpCost;
                                double cost = TA.GetCost(tech, veh, data, this._settings, year, this._yearOfInitialApplication[i][tech.Index],
                                    this._startYear, this._endYear, this._techGroup, techKey, out erpCost, ref this._committedERPCosts);
                                impact.Cost   [j] += cost;
                                impact.ERPCost[j] += erpCost;
                                impact.MaintenanceCost[j] += TA.GetMaintenanceCost(tech, veh, year);
                                impact.RepairCost     [j] += TA.GetRepairCost     (tech, veh, year);
                                double fc, fcg, fcgShare, fcNoSyn, fcgNoSyn;
                                fcNoSyn = TA.GetImprovement(tech, techClass, vmd.TechUsed, vmd.TechSuperseded, this._techGroup, techKey,
                                    true, out fcgNoSyn, out fcgShare);
                                fc = TA.GetImprovement(tech, techClass, vmd.TechUsed, vmd.TechSuperseded, this._techGroup, techKey,
                                    false, out fcg, out fcgShare);
                                if      (veh.IsRCAuto ) { impact.OffCycleCredit[j] += tech.Attributes[veh.TechnologyClass].OffCycleCreditPC; }
                                else if (veh.IsRCTruck) { impact.OffCycleCredit[j] += tech.Attributes[veh.TechnologyClass].OffCycleCreditLT; }
                                FuelType primaryFuel = impact.NewFuelShare[j].PrimaryFuel;
                                if (TI.IsDieselization(tech.Index))
                                {   
                                    impact.NewFuelEconomy [j][FuelType.Diesel] = impact.NewFuelEconomy [j].Gasoline / (1 - fc);
                                    impact.NewFuelEconomy2[j][FuelType.Diesel] = impact.NewFuelEconomy2[j].Gasoline / (1 - fcNoSyn);
                                    impact.NewFuelEconomy [j][FuelType.Gasoline ] = 0;
                                    impact.NewFuelEconomy [j][FuelType.Ethanol85] = 0;
                                    impact.NewFuelEconomy2[j][FuelType.Gasoline ] = 0;
                                    impact.NewFuelEconomy2[j][FuelType.Ethanol85] = 0;
                                    impact.NewFuelShare[j][FuelType.Gasoline ] = 0.0;
                                    impact.NewFuelShare[j][FuelType.Ethanol85] = 0.0;
                                    impact.NewFuelShare[j][FuelType.Diesel   ] = 1.0;
                                }
                                else if (tech.Index == TI.SHEV1 || (tech.Index == TI.SHEV2 && !shev1Used))
                                {   
                                    impact.NewFuelEconomy[j] = impact.NewFuelEconomy2[j];
                                    impact.NewFuelEconomy [j][primaryFuel] = impact.NewFuelEconomy [j][primaryFuel] / (1 - fc);
                                    impact.NewFuelEconomy2[j][primaryFuel] = impact.NewFuelEconomy2[j][primaryFuel] / (1 - fcNoSyn);
                                    impact.NewFuelEconomy [j][FuelType.Ethanol85] = 0;
                                    impact.NewFuelEconomy2[j][FuelType.Ethanol85] = 0;
                                    impact.NewFuelShare[j][FuelType.Gasoline ] = 1.0;
                                    impact.NewFuelShare[j][FuelType.Ethanol85] = 0.0;
                                }
                                else if (TI.IsPartialElectric(tech.Index))
                                {   
                                    double newFE1 = Standards.GetVehicleFuelEconomy(impact.NewFuelEconomy [j], impact.NewFuelShare[j], 1) / (1 - fc);
                                    double newFE2 = Standards.GetVehicleFuelEconomy(impact.NewFuelEconomy2[j], impact.NewFuelShare[j], 1) / (1 - fcNoSyn);
                                    impact.NewFuelEconomy [j][primaryFuel] = impact.NewFuelEconomy [j][primaryFuel] / (1 - fcg);
                                    impact.NewFuelEconomy2[j][primaryFuel] = impact.NewFuelEconomy2[j][primaryFuel] / (1 - fcgNoSyn);
                                    impact.NewFuelEconomy [j][FuelType.Electricity] = (1 - fcgShare) / (1 / newFE1 - fcgShare / impact.NewFuelEconomy [j][primaryFuel]);
                                    impact.NewFuelEconomy2[j][FuelType.Electricity] = (1 - fcgShare) / (1 / newFE2 - fcgShare / impact.NewFuelEconomy2[j][primaryFuel]);
                                    impact.NewFuelEconomy [j][FuelType.Ethanol85] = 0;
                                    impact.NewFuelEconomy2[j][FuelType.Ethanol85] = 0;
                                    impact.NewFuelShare[j][primaryFuel         ] = fcgShare;
                                    impact.NewFuelShare[j][FuelType.Electricity] = 1 - fcgShare;
                                    impact.NewFuelShare[j][FuelType.Ethanol85  ] = 0.0;
                                }
                                else if (tech.Index == TI.EV1)
                                {   
                                    double curFE1 = Standards.GetVehicleFuelEconomy(impact.NewFuelEconomy [j], impact.NewFuelShare[j], 1);
                                    double curFE2 = Standards.GetVehicleFuelEconomy(impact.NewFuelEconomy2[j], impact.NewFuelShare[j], 1);
                                    impact.NewFuelEconomy [j][FuelType.Electricity] = curFE1 / (1 - fc);
                                    impact.NewFuelEconomy2[j][FuelType.Electricity] = curFE2 / (1 - fcNoSyn);
                                    impact.NewFuelEconomy [j][FuelType.Gasoline] = 0;
                                    impact.NewFuelEconomy [j][FuelType.Diesel  ] = 0;
                                    impact.NewFuelEconomy2[j][FuelType.Gasoline] = 0;
                                    impact.NewFuelEconomy2[j][FuelType.Diesel  ] = 0;
                                    impact.NewFuelShare[j][FuelType.Gasoline   ] = 0.0;
                                    impact.NewFuelShare[j][FuelType.Diesel     ] = 0.0;
                                    impact.NewFuelShare[j][FuelType.Electricity] = 1.0;
                                }
                                else if (tech.Index == TI.FCV)
                                {   
                                    double curFE1 = Standards.GetVehicleFuelEconomy(impact.NewFuelEconomy [j], impact.NewFuelShare[j], 1);
                                    double curFE2 = Standards.GetVehicleFuelEconomy(impact.NewFuelEconomy2[j], impact.NewFuelShare[j], 1);
                                    impact.NewFuelEconomy [j][FuelType.Hydrogen] = curFE1 / (1 - fc);
                                    impact.NewFuelEconomy2[j][FuelType.Hydrogen] = curFE2 / (1 - fcNoSyn);
                                    impact.NewFuelEconomy [j][FuelType.Gasoline   ] = 0;
                                    impact.NewFuelEconomy [j][FuelType.Diesel     ] = 0;
                                    impact.NewFuelEconomy [j][FuelType.Electricity] = 0;
                                    impact.NewFuelEconomy2[j][FuelType.Gasoline   ] = 0;
                                    impact.NewFuelEconomy2[j][FuelType.Diesel     ] = 0;
                                    impact.NewFuelEconomy2[j][FuelType.Electricity] = 0;
                                    impact.NewFuelShare[j][FuelType.Gasoline   ] = 0.0;
                                    impact.NewFuelShare[j][FuelType.Diesel     ] = 0.0;
                                    impact.NewFuelShare[j][FuelType.Electricity] = 0.0;
                                    impact.NewFuelShare[j][FuelType.Hydrogen   ] = 1.0;
                                }
                                else
                                {   
                                    if (impact.NewFuelShare[j].Gasoline > 0)
                                    {
                                        impact.NewFuelEconomy [j][FuelType.Gasoline] = impact.NewFuelEconomy [j].Gasoline / (1 - fc);
                                        impact.NewFuelEconomy2[j][FuelType.Gasoline] = impact.NewFuelEconomy2[j].Gasoline / (1 - fcNoSyn);
                                    }
                                    if (impact.NewFuelShare[j].Diesel > 0)
                                    {
                                        impact.NewFuelEconomy [j][FuelType.Diesel] = impact.NewFuelEconomy [j].Diesel / (1 - fc);
                                        impact.NewFuelEconomy2[j][FuelType.Diesel] = impact.NewFuelEconomy2[j].Diesel / (1 - fcNoSyn);
                                    }
                                    if (impact.NewFuelShare[j].Ethanol85 > 0)
                                    {
                                        impact.NewFuelEconomy [j][FuelType.Ethanol85] = impact.NewFuelEconomy [j].Ethanol85 / (1 - fc);
                                        impact.NewFuelEconomy2[j][FuelType.Ethanol85] = impact.NewFuelEconomy2[j].Ethanol85 / (1 - fcNoSyn);
                                    }
                                    if (impact.NewFuelShare[j].Electricity > 0)
                                    {
                                        impact.NewFuelEconomy [j][FuelType.Electricity] = impact.NewFuelEconomy [j].Electricity / (1 - fc);
                                        impact.NewFuelEconomy2[j][FuelType.Electricity] = impact.NewFuelEconomy2[j].Electricity / (1 - fcNoSyn);
                                    }
                                    if (impact.NewFuelShare[j].Hydrogen > 0)
                                    {
                                        impact.NewFuelEconomy [j][FuelType.Hydrogen] = impact.NewFuelEconomy [j].Hydrogen / (1 - fc);
                                        impact.NewFuelEconomy2[j][FuelType.Hydrogen] = impact.NewFuelEconomy2[j].Hydrogen / (1 - fcNoSyn);
                                    }
                                }
                                TechnologyAttributesTable.TechnologyAttributes attr = tech.Attributes[techClass];
                                if      (attr.DeltaWeightPCT != 0) { impact.DeltaCurbWeight[j] -= (vd.BaseWeight * attr.DeltaWeightPCT); }
                                else if (attr.DeltaWeightLBS != 0) { impact.DeltaCurbWeight[j] -= (attr.DeltaWeightLBS); }
                            } 
                        } 
                    } 
                } 
                if (!impact.Ignore)
                {   
                    for (int j = lingerYrIndex; j <= this._yrIndex; j++)
                    {
                        if (!impact.HasImpact[j]) { continue; }
                        FuelValue vehFS  = impact.Vehicle[j].Description.FuelShare;
                        FuelValue vehFE1 = impact.Vehicle[j].Description.FuelEconomy;
                        FuelValue vehFE2 = impact.Vehicle[j].Description.FuelEconomy2;
                        double vehFE1Combined = Standards.GetVehicleFuelEconomy(vehFE1, vehFS, 1);
                        double vehFE2Combined = Standards.GetVehicleFuelEconomy(vehFE2, vehFS, 1);
                        impact.Improvement [j][FuelType.Gasoline   ] = (impact.NewFuelEconomy [j].Gasoline    == 0) ? 0 : ((vehFE1.Gasoline    == 0) ? vehFE1Combined : vehFE1.Gasoline   ) / impact.NewFuelEconomy [j].Gasoline;
                        impact.Improvement [j][FuelType.Diesel     ] = (impact.NewFuelEconomy [j].Diesel      == 0) ? 0 : ((vehFE1.Diesel      == 0) ? vehFE1Combined : vehFE1.Diesel     ) / impact.NewFuelEconomy [j].Diesel;
                        impact.Improvement [j][FuelType.Ethanol85  ] = (impact.NewFuelEconomy [j].Ethanol85   == 0) ? 0 : ((vehFE1.Ethanol85   == 0) ? vehFE1Combined : vehFE1.Ethanol85  ) / impact.NewFuelEconomy [j].Ethanol85;
                        impact.Improvement [j][FuelType.Electricity] = (impact.NewFuelEconomy [j].Electricity == 0) ? 0 : ((vehFE1.Electricity == 0) ? vehFE1Combined : vehFE1.Electricity) / impact.NewFuelEconomy [j].Electricity;
                        impact.Improvement [j][FuelType.Hydrogen   ] = (impact.NewFuelEconomy [j].Hydrogen    == 0) ? 0 : ((vehFE1.Hydrogen    == 0) ? vehFE1Combined : vehFE1.Hydrogen   ) / impact.NewFuelEconomy [j].Hydrogen;
                        impact.Improvement2[j][FuelType.Gasoline   ] = (impact.NewFuelEconomy2[j].Gasoline    == 0) ? 0 : ((vehFE2.Gasoline    == 0) ? vehFE2Combined : vehFE2.Gasoline   ) / impact.NewFuelEconomy2[j].Gasoline;
                        impact.Improvement2[j][FuelType.Diesel     ] = (impact.NewFuelEconomy2[j].Diesel      == 0) ? 0 : ((vehFE2.Diesel      == 0) ? vehFE2Combined : vehFE2.Diesel     ) / impact.NewFuelEconomy2[j].Diesel;
                        impact.Improvement2[j][FuelType.Ethanol85  ] = (impact.NewFuelEconomy2[j].Ethanol85   == 0) ? 0 : ((vehFE2.Ethanol85   == 0) ? vehFE2Combined : vehFE2.Ethanol85  ) / impact.NewFuelEconomy2[j].Ethanol85;
                        impact.Improvement2[j][FuelType.Electricity] = (impact.NewFuelEconomy2[j].Electricity == 0) ? 0 : ((vehFE2.Electricity == 0) ? vehFE2Combined : vehFE2.Electricity) / impact.NewFuelEconomy2[j].Electricity;
                        impact.Improvement2[j][FuelType.Hydrogen   ] = (impact.NewFuelEconomy2[j].Hydrogen    == 0) ? 0 : ((vehFE2.Hydrogen    == 0) ? vehFE2Combined : vehFE2.Hydrogen   ) / impact.NewFuelEconomy2[j].Hydrogen;
                        impact.DeltaFuelEconomy [j][FuelType.Gasoline   ] = impact.NewFuelEconomy [j].Gasoline    - vehFE1.Gasoline;
                        impact.DeltaFuelEconomy [j][FuelType.Diesel     ] = impact.NewFuelEconomy [j].Diesel      - vehFE1.Diesel;
                        impact.DeltaFuelEconomy [j][FuelType.Ethanol85  ] = impact.NewFuelEconomy [j].Ethanol85   - vehFE1.Ethanol85;
                        impact.DeltaFuelEconomy [j][FuelType.Electricity] = impact.NewFuelEconomy [j].Electricity - vehFE1.Electricity;
                        impact.DeltaFuelEconomy [j][FuelType.Hydrogen   ] = impact.NewFuelEconomy [j].Hydrogen    - vehFE1.Hydrogen;
                        impact.DeltaFuelEconomy2[j][FuelType.Gasoline   ] = impact.NewFuelEconomy2[j].Gasoline    - vehFE2.Gasoline;
                        impact.DeltaFuelEconomy2[j][FuelType.Diesel     ] = impact.NewFuelEconomy2[j].Diesel      - vehFE2.Diesel;
                        impact.DeltaFuelEconomy2[j][FuelType.Ethanol85  ] = impact.NewFuelEconomy2[j].Ethanol85   - vehFE2.Ethanol85;
                        impact.DeltaFuelEconomy2[j][FuelType.Electricity] = impact.NewFuelEconomy2[j].Electricity - vehFE2.Electricity;
                        impact.DeltaFuelEconomy2[j][FuelType.Hydrogen   ] = impact.NewFuelEconomy2[j].Hydrogen    - vehFE2.Hydrogen;
                    } 
                } 
            } 
        }
        public void IgnoreTechnologies()
        {
            for (int i = 0; i < this._vehCount; i++)
            {
                if (this._ignore[i] == 0)
                {   
                    Vehicle.CModelData vmd = this._vehs[i].ModelData;
                    for (int j = 0; j < this._techCount; j++)
                    {   
                        if ((this._techKeys[i] & TA.TechKeys[j]) != 0)
                        {
                            vmd.TechIgnored[this._techGroup[j].Index] = true;
                        }
                    } 
                } 
            } 
        }
        #endregion
        #region 
        public void ApplyFinding()
        {
            if (!this.IsValid  ) { throw new InvalidOperationException("The ComplianceFinding being applied is not valid."); }
            if (this._isApplied) { throw new InvalidOperationException("The ComplianceFinding has already been applied."  ); }
            this._isApplied = true;
            this.LogFinding();
            this.ApplyFinding_CheckSplitCf();
            for (int i = 0; i < this._vehCount; i++)
            {
                if (this._ignore[i] == 0)
                {   
                    this.ApplyTechImpact(i, this._techImpacts[i], this._techKeys[i], this._techLingerKeys[i]);
                }
            }
        }
        void ApplyTechImpact(int vehIndex, TechImpact impact, ulong curTechKey, ulong lingerTechKey)
        {
            for (int i = 0; i <= this._yrIndex; i++)
            {
                ModelYear modelYear = new ModelYear(ModelYear.MinYear + i);
                if (impact.HasImpact[i])
                {   
                    Vehicle                   veh = impact.Vehicle[i];
                    Vehicle     .CDescription vd  = veh.Description;
                    Vehicle     .CModelData   vmd = veh.ModelData;
                    Manufacturer.CModelData   mmd = veh.Manufacturer.ModelData;
                    ulong techKey = (i == this._yrIndex) ? curTechKey : lingerTechKey;
                    RegulatoryClass regClass = veh.RegClass;
                    double          vehSales = vd.Sales[i];
                    ulong baseTech = 0;
                    for (int j = this._techCount - 1; j >= 0; j--)
                    {   
                        if ((techKey & TA.TechKeys[j]) != 0) { baseTech = TA.TechKeys[j]; break; }
                    }
                    for (int j = 0; j < this._techCount; j++)
                    {   
                        if ((techKey & TA.TechKeys[j]) != 0)
                        {   
                            int techIndex = this._techGroup[j].Index;
                            this.ApplyTechnology(veh, vmd, mmd, regClass, vehSales, this._techGroup[j], modelYear,
                                this._yearOfInitialApplication[vehIndex][techIndex], TA.TechKeys[j] != baseTech);
                        }
                    } 
                    if (impact.DeltaCurbWeight[i] != 0) { vd.CurbWeight += impact.DeltaCurbWeight[i]; }
                    if (!impact.NewFuelShare[i].IsValid || !impact.NewFuelEconomy[i].IsValid || !impact.NewFuelEconomy2[i].IsValid)
                    {
                        throw new InvalidOperationException(
                            "Error in ComplianceFinding.  Vehicle's new Fuel Economy and/or Fuel Share values are not valid.");
                    }
                    else
                    {
                        vd.FuelShare    = impact.NewFuelShare   [i];
                        vd.FuelEconomy  = impact.NewFuelEconomy [i];
                        vd.FuelEconomy2 = impact.NewFuelEconomy2[i];
                    }
                    vmd.OffCycleCredit += impact.OffCycleCredit[i];
                    vmd.TechCost                  +=             impact.Cost           [i] ;
                    vmd.LossOfValue               +=             impact.LossOfValue    [i] ;
                    vmd.MaintenanceCost           +=             impact.MaintenanceCost[i] ;
                    vmd.RepairCost                +=             impact.RepairCost     [i] ;
                    mmd.TechCost       [regClass] += (vehSales * impact.Cost           [i]);
                    mmd.LossOfValue    [regClass] += (vehSales * impact.LossOfValue    [i]);
                    mmd.MaintenanceCost[regClass] += (vehSales * impact.MaintenanceCost[i]);
                    mmd.RepairCost     [regClass] += (vehSales * impact.RepairCost     [i]);
                } 
            } 
        }
        void ApplyTechnology(Vehicle veh, Vehicle.CModelData vmd, Manufacturer.CModelData mmd, RegulatoryClass regClass,
            double vehSales, Technology tech, ModelYear year, int yearOfInitialApplication, bool backfilling)
        {
            TA.InstallTechnologyOnVehicle(veh, tech, year, yearOfInitialApplication, this._settings, backfilling,
                mmd.TechExhausted[tech.Index][regClass]);
            mmd.TechUsedSales   [tech.Index][regClass] += vehSales;
            mmd.TechAppliedSales[tech.Index][regClass] += vehSales;
            if (!this._settings.OperatingModes.IgnorePhaseIn)
            {   
                TA.CheckPhaseIn(this._settings, mmd, regClass, tech, year, this._startYear - ModelYear.MinYear);
            }
        }
        void ApplyFinding_CheckSplitCf()
        {
            if (!this._isSplitCf) { return; }
            if (!this._isEngOrTrn)
            {
                throw new Exception("The ComplianceFinding was marked as split, however, the technologies selected " +
                    "for applicability are not engine or transmission type.");
            }
            List<Vehicle>[] splitVehs = new List<Vehicle>[this._yrIndex + 1];
            for (int i = 0; i < this._vehCount; i++)
            {
                TechImpact impact = this._techImpacts[i];
                if (impact.Ignore) { continue; }
                for (int j = 0; j <= this._yrIndex; j++)
                {
                    if (impact.HasImpact[j])
                    {   
                        if (splitVehs[j] == null) { splitVehs[j] = new List<Vehicle>(); }
                        splitVehs[j].Add(impact.Vehicle[j]);
                    }
                } 
            } 
            int code = -1;
            for (int i = this._yrIndex; i >= 0; i--)
            {
                if (splitVehs[i] != null)
                {   
                    if (this._techType == TechnologyType.Engine)
                    {   
                        Engine eng = splitVehs[i][0].Engine.Split(splitVehs[i], code);
                        splitVehs[i][0].Manufacturer.Engines.Add(eng);
                        code = eng.Description.Code;
                    }
                    else 
                    {   
                        Transmission trn = splitVehs[i][0].Transmission.Split(splitVehs[i], code);
                        splitVehs[i][0].Manufacturer.Transmissions.Add(trn);
                        code = trn.Description.Code;
                    }
                } 
            } 
        }
        #endregion
        #region 
        public void LogFinding()
        {
            if (!this.IsValid || this._logWriter.IsEmpty) { return; }
            Manufacturer mfr = this._modelYearData[this._yrIndex].Manufacturers[this._mfrIndex];
            ComplianceWriter log    = (this._logWriter.CFs    == null) ? null : this._logWriter.CFs   [this._scen.Index];
            ComplianceWriter extLog = (this._logWriter.ExtCFs == null) ? null : this._logWriter.ExtCFs[this._scen.Index];
            if (log    != null && this._isApplied) { this.LogFinding(log   , mfr.ModelData); }
            if (extLog != null                   ) { this.LogFinding(extLog, mfr.ModelData); }
        }
        void LogFinding(ComplianceWriter cw, Manufacturer.CModelData mmd)
        {
            StringBuilder line = new StringBuilder();
            int lineNumber = cw.LineCount + 1;
            for (int i = 0; i < this._vehCount; i++)
            {
                TechImpact impact = this._techImpacts[i];
                if (!impact.Ignore)
                {   
                    Vehicle baseVeh     = impact.Vehicle[this._yrIndex];
                    int     baseVehCode = baseVeh.Description.Code;
                    for (int j = impact.HasImpact.Length - 1; j >= 0; j--)
                    {   
                        if (impact.HasImpact[j] == false) { continue; }
                        Vehicle              veh = impact.Vehicle[j];
                        Vehicle.CDescription vd  = veh.Description;
                        ulong techKey = (j == this._yrIndex) ? this._techKeys[i] : this._techLingerKeys[i];
                        line.Append(lineNumber++                ); line.Append(TAB);
                        line.Append(this._index                 ); line.Append(TAB);
                        line.Append(this._scen.Index            ); line.Append(TAB);
                        line.Append(vd.Manufacturer             ); line.Append(TAB);
                        if (this._isApplied) { line.Append((this._overcomply) ? 'O' : 'Y'); } line.Append(TAB);
                        line.Append(this._efficiency            ); line.Append(TAB);
                        line.Append(this._totalDeltaRPE     .ToString("0.####")); line.Append(TAB);
                        line.Append(this._totalDeltaFines   .ToString("0.####")); line.Append(TAB);
                        line.Append(this._totalFuelSavings  .ToString("0.####")); line.Append(TAB);
                        line.Append(this._totalAffectedSales.ToString("0.####")); line.Append(TAB);
                        RegulatoryClass regClass = veh.RegClass;
                        line.Append(               regClass ); line.Append(TAB);
                        line.Append(mmd.Standard  [regClass]); line.Append(TAB);
                        line.Append(this._curCafe [regClass]); line.Append(TAB);
                        line.Append(this._newCafe [regClass]); line.Append(TAB);
                        line.Append(this._curFines[regClass]); line.Append(TAB);
                        line.Append(this._newFines[regClass]); line.Append(TAB);
                        if (baseVehCode != vd.Code)
                        {   
                            line.Append('['); line.Append(vd.Code); line.Append(']');
                        }
                        line.Append(vd.Code); line.Append(TAB);
                        bool isFirstTech = true;
                        for (int k = 0; k < this._techCount; k++)
                        {
                            if ((techKey & TA.TechKeys[k]) != 0)
                            {   
                                if (isFirstTech) { isFirstTech = false; } 
                                else { line.Append(", "); }
                                line.Append(this._techGroup[k].Abbr);
                            } 
                        } 
                        line.Append(TAB);
                        line.Append(j + ModelYear.MinYear);                          line.Append(TAB);
                        line.Append(vd.FuelEconomy             .ToString("0.####")); line.Append(TAB);
                        line.Append(vd.FuelShare               .ToString("0.####")); line.Append(TAB);
                        line.Append(impact.NewFuelEconomy   [j].ToString("0.####")); line.Append(TAB);
                        line.Append(impact.NewFuelShare     [j].ToString("0.####")); line.Append(TAB);
                        line.Append(vd.Sales                [j].ToString(        )); line.Append(TAB);
                        line.Append(impact.Cost             [j].ToString("0.####")); line.Append(TAB);
                        line.Append(impact.ERPCost          [j].ToString("0.####")); line.Append(TAB);
                        line.Append(impact.Improvement      [j].ToString("0.####")); line.Append(TAB);
                        line.Append(impact.DeltaFuelEconomy [j].ToString("0.####")); line.Append(TAB);
                        line.Append(impact.OffCycleCredit   [j].ToString("0.####")); line.Append(TAB);
                        line.Append(impact.DeltaCurbWeight  [j].ToString("0.####")); line.Append(TAB);
                        line.Append(impact.NewFuelEconomy2  [j].ToString("0.####")); line.Append(TAB);
                        line.Append(impact.Improvement2     [j].ToString("0.####")); line.Append(TAB);
                        line.Append(impact.DeltaFuelEconomy2[j].ToString("0.####"));
                        line.Append(NewLine);
                    } 
                } 
            }
            cw.Write(line.ToString());
        }
        #endregion
        #endregion
        #region 
        public int Index { get { return this._index; } }
        public bool IsValid { get { return !double.IsNaN(this._efficiency); } }
        public bool IsEfficient { get { return (this._efficiency < 0); } }
        public bool InCompliance { get { return this.IsValid && (this._effectiveNewFines.TotalAll <= 0); } }
        public bool IsApplied { get { return this._isApplied; } }
        public double Efficiency { get { return this._efficiency; } }
        public double TechCost { get { return this._techCost; } }
        #endregion
        #region 
        const string TAB     = "\t";
        const string TAB_8   = "\t\t\t\t\t\t\t\t";
        const string TAB_14  = "\t\t\t\t\t\t\t\t\t\t\t\t\t\t";
        const string NewLine = "\n";
        Scenario             _scen;
        ModelYear            _year;
        int                  _yrIndex;          
        int                  _yrYear;           
        int                  _startYear;
        int                  _endYear;
        Industry[]           _modelYearData;
        int                  _mfrIndex;
        ModelingSettings     _settings;
        Parameters           _parameters;       
        bool                 _multiYear;
        bool                 _ignorePhaseIn;
        bool                 _backfill;
        bool                 _splitShared;
        bool                 _splitMixed;
        bool                 _skipShared;
        bool                 _skipMixed;
        LogWriter            _logWriter;
        bool                 _isEngOrTrn;       
        TechnologyType       _techType;         
        List<Technology>     _techGroup;        
        int                  _techCount;
        internal int _index;                    
        internal int _vehCount;                 
        internal int _capacity;                 
        internal Vehicle   []    _vehs;         
        internal int       []    _ignore;       
        int                   _techIndex;       
        internal ulong     [] _techKeys;        
        internal ulong     [] _techLingerKeys;  
        int                [] _lingerYear;      
        internal TechImpact[] _techImpacts;     
        int              [][] _yearOfInitialApplication;
        bool[] _committedERPCosts;
        internal bool _hasVehs;
        internal bool _isRcMix;                 
        internal bool _isEnablesMix;            
        internal bool _isSplitCf;
        internal bool _isApplied;
        internal bool _overcomply;
        internal double _efficiency;
        internal double _techCost;
        internal RCDouble _newCafe;
        RCDouble _newFines;
        RCDouble _effectiveNewFines;
        RCDouble _curCafe;
        RCDouble _curFines;
        double _totalDeltaRPE;
        double _totalDeltaFines;
        double _totalFuelSavings;
        double _totalAffectedSales;
        #endregion
    }
}

