#region << Using Directives >>
using System;
using System.Collections;
using System.Collections.Generic;
using Volpe.Cafe;
using Volpe.Cafe.Data;
using Volpe.Cafe.Settings;
using RC = Volpe.Cafe.RegulatoryClass;
using VT = Volpe.Cafe.VehicleType;
#endregion
namespace Volpe.Cafe.Model
{
    public sealed class Standards
    {
        #region 
        #region 
        public static RCDouble GetIndustryStandard(Industry data, Scenario scen, ModelYear year, bool calcUnregulated)
        {
            RCDouble standard = new RCDouble();
            standard.DomesticAuto = Standards.GetIndustryStandard(data, scen, year, RC.DomesticAuto);
            standard.ImportedAuto = Standards.GetIndustryStandard(data, scen, year, RC.ImportedAuto);
            standard.LightTruck   = Standards.GetIndustryStandard(data, scen, year, RC.LightTruck);
            if (calcUnregulated) { standard.Unregulated = Standards.GetIndustryStandard(data, scen, year, RC.Unregulated); }
            return standard;
        }
        public static double GetIndustryStandard(Industry data, Scenario scen, ModelYear year, RC regClass)
        {
            List<Manufacturer> mfrs = data.Manufacturers;
            double standardSum = 0, salesSum    = 0;
            for (int i = 0, mfrCount = mfrs.Count; i < mfrCount; i++)
            {   
                Manufacturer manufacturer = mfrs[i];
                double mfrSales = Standards.GetSales   (manufacturer,       year, regClass);
                double mfrStnd  = Standards.GetStandard(manufacturer, scen, year, regClass, null, 0);
                standardSum += (mfrSales == 0 || mfrStnd == 0) ? 0 : mfrSales / mfrStnd;
                salesSum    += mfrSales;
            }
            return salesSum / standardSum;
        }
        public static void CalcStandard(Manufacturer mfr, Scenario scen, ModelYear year)
        {
            Standards.CalcStandard(mfr, scen, year, RC.DomesticAuto);
            Standards.CalcStandard(mfr, scen, year, RC.ImportedAuto);
            Standards.CalcStandard(mfr, scen, year, RC.LightTruck  );
            Standards.CalcStandard(mfr, scen, year, RC.Unregulated );
        }
        public static void CalcStandard(Manufacturer mfr, Scenario scen, ModelYear year, RC regClass)
        {
            mfr.ModelData.Standard[regClass] = Standards.GetStandard(mfr, scen, year, regClass, null, 0);
        }
        public static double GetStandard(Manufacturer mfr, Scenario scen, ModelYear year, RC regClass, TechImpact[] techImpacts,
            int techImpactCount)
        {
            if (techImpactCount > 0 && (techImpacts == null || techImpacts.Length < techImpactCount))
            {   
                throw new ArgumentOutOfRangeException("techImpacts", "techImpacts is null or less than techImpactCount.");
            }
            Manufacturer.CModelData mmd = mfr.ModelData;
            double standard = 0;
            int yrIndex = year.Index;
            int fncType;
            double[,] fncCoeff;
            int numCoeff = Standards.GetCoefficients(regClass, scen, year, out fncType, out fncCoeff);
            if (fncType == 0 || fncCoeff == null || regClass == RC.Unregulated)
            {   
                standard = 0;
            }
            else if (fncType == 1)
            {   
                standard = fncCoeff[0, yrIndex];
            }
            else if (fncType == 99)
            {   
                standard = mmd.DefaultStandard[yrIndex][regClass];
            }
            else
            {   
                double prevStandard = mmd.PreliminaryStandard[regClass];
                bool noPrevStandard = (prevStandard == 0 || double.IsNaN(prevStandard));
                bool areaSystem     = FunctionInformation.IsAreaBasedFunction(fncType);
                bool areaChanged    = noPrevStandard, weightChanged = noPrevStandard;
                if (!noPrevStandard)
                {
                    if (areaSystem)
                    {   
                    }
                    else
                    {   
                        for (int i = 0; i < techImpactCount; i++)
                        {
                            if (!techImpacts[i].Ignore && techImpacts[i].DeltaCurbWeight[yrIndex] != 0) { weightChanged = true; break; }
                        }
                    }
                }
                if (areaChanged || weightChanged)
                {   
                    double sales  = 0;      
                    double stdSum = 0;      
                    List<Vehicle> vehs = mfr.Vehicles;
                    for (int i = 0, vehCount = vehs.Count; i < vehCount; i++)
                    {
                        Vehicle veh = vehs[i];
                        if (veh.RegClass == regClass)
                        {
                            Vehicle.CDescription vd = veh.Description;
                            double target = 0;
                            int idx = -1;
                            for (int j = 0; j < techImpactCount; j++)
                            {
                                if (!techImpacts[j].Ignore && veh == techImpacts[j].Vehicle[yrIndex]) { idx = j; break; }
                            }
                            target = Standards.GetTarget(veh, scen, year, (idx == -1) ? double.NaN : vd.Footprint,
                                (idx == -1) ? double.NaN : vd.CurbWeight + techImpacts[idx].DeltaCurbWeight[yrIndex], fncType, fncCoeff);
                            sales  += vd.Sales[yrIndex];
                            stdSum += vd.Sales[yrIndex] * target;
                        } 
                    } 
                    standard = sales / stdSum;
                }
                else
                {   
                    standard = prevStandard;
                }
            }
            mmd.PreliminaryStandard[regClass] = Math.Round(standard, 1);
            if (regClass == RC.LightTruck)
            {
                if (standard < scen.LtMinStndMpg[yrIndex]) { standard = scen.LtMinStndMpg[yrIndex]; }
                double pctIndStnd = scen.LtMinStndPct[yrIndex] * mmd.AverageStandard.LightTruck;
                if (standard < pctIndStnd) { standard = pctIndStnd; }
            }
            else if (regClass == RC.DomesticAuto || regClass == RC.ImportedAuto)
            {
                if (standard < scen.AutoMinStndMpg[yrIndex]) { standard = scen.AutoMinStndMpg[yrIndex]; }
                double pctIndStnd = scen.AutoMinStndPct[yrIndex] * mmd.AverageStandard[regClass];
                if (standard < pctIndStnd) { standard = pctIndStnd; }
            }
            return Math.Round(standard, 1);
        }
        public static double GetTarget(Vehicle vehicle, Scenario scen, ModelYear year)
        {
            return Standards.GetTarget(vehicle, scen, year, double.NaN, double.NaN, 0, null);
        }
        public static double GetTarget(Vehicle vehicle, Scenario scen, ModelYear year, double area, double weight, int fncType,
            double[,] fncCoeff)
        {
            if (double.IsNaN(area  ) || area   <= 0) { area   = vehicle.Description.Footprint;  }
            if (double.IsNaN(weight) || weight <= 0) { weight = vehicle.Description.CurbWeight; }
            if (fncType == 0 || fncCoeff == null)
            {   
                Standards.GetCoefficients(vehicle.RegClass, scen, year, out fncType, out fncCoeff);
            }
            return Standards.GetTarget(area, weight, fncType, fncCoeff, year.Index);
        }
        public static double GetTarget(double area, double weight, int fncType, double[] fncCoeff)
        {
            return Standards.GetTarget(area, weight, fncType,
                fncCoeff[0], fncCoeff[1], fncCoeff[2], fncCoeff[3], fncCoeff[4], fncCoeff[5], fncCoeff[6], fncCoeff[7]);
        }
        public static double GetTarget(double area, double weight, int fncType, double[,] fncCoeff, int yrIndex)
        {
            return Standards.GetTarget(area, weight, fncType,
                fncCoeff[0, yrIndex], fncCoeff[1, yrIndex], fncCoeff[2, yrIndex], fncCoeff[3, yrIndex],
                fncCoeff[4, yrIndex], fncCoeff[5, yrIndex], fncCoeff[6, yrIndex], fncCoeff[7, yrIndex]);
        }
        public static double GetTarget(double area, double weight, int fncType, double A, double B, double C, double D,
            double E, double F, double G, double H)
        {
            double target = 0.0, invA = 1.0 / A, invB = 1.0 / B;
            double exp;
            switch (fncType)
            {
                case 1:
                    target = invA;
                    break;
                case 2:     
                case 4:     
                    exp    = Math.Exp((((fncType == 2) ? area : weight) - C) / D);
                    target = invA + (invB - invA) * (exp / (1.0 + exp));
                    break;
                case 3:     
                case 5:     
                    exp    = Math.Exp(1.0 - ((fncType == 3) ? area : weight) / C);
                    target = invA - exp * invB;
                    break;
                case 6:     
                case 7:     
                    target = C * ((fncType == 6) ? area : weight) + D;                  
                    target = (target < invA) ? invA : (target > invB) ? invB : target;  
                    break;
                case 8:     
                case 9:     
                    target = C * ((fncType == 8) ? area : weight) + D;                  
                    target = (target > invB) ? invB : target;                           
                    break;
                case 206:   
                case 207:   
                    double t1 = C * ((fncType == 206) ? area : weight) + D;             
                    t1 = (t1 < invA) ? invA : (t1 > invB) ? invB : t1;                  
                    double invE = 1.0 / E, invF = 1.0 / F;
                    double t2 = G * ((fncType == 206) ? area : weight) + H;             
                    t2 = (t2 < invE) ? invE : (t2 > invF) ? invF : t2;                  
                    target = (t1 < t2) ? t1 : t2;
                    break;
            }
            return target;
        }
        public static RC GetRegClass(Vehicle vehicle, Scenario scen)
        {
            VT vehType = vehicle.VehicleType;
            if (vehType == VT.None) { return RC.Unregulated; }
            else
            {   
                string merging = scen.Options.RegulatoryMerging.ToUpper();
                bool   isRegLt = (vehicle.Description.RegulatoryIndicator == "LT"); 
                if      (isRegLt                                || merging == "RC3") { return RC.LightTruck;   }
                else if ((vehType & VT.Domestic) == VT.Domestic || merging == "RC1") { return RC.DomesticAuto; }
                else if ((vehType & VT.Imported) == VT.Imported                    ) { return RC.ImportedAuto; }
                else                                                                 { return RC.Unregulated;  }
            }
        }
        #endregion
        #region 
        public static RCDouble GetSales(Manufacturer mfr, ModelYear year)
        {
            RCDouble sales = new RCDouble();
            sales.DomesticAuto = Standards.GetSales(mfr, year, RC.DomesticAuto);
            sales.ImportedAuto = Standards.GetSales(mfr, year, RC.ImportedAuto);
            sales.LightTruck   = Standards.GetSales(mfr, year, RC.LightTruck  );
            sales.Unregulated  = Standards.GetSales(mfr, year, RC.Unregulated );
            return sales;
        }
        public static double GetSales(Manufacturer mfr, ModelYear year, RC regClass)
        {
            return Standards.GetSales(mfr.Vehicles, year, regClass);
        }
        public static double GetSales(List<Vehicle> vehs, ModelYear year, RC regClass)
        {
            double sales = 0;
            int yrIndex = year.Index;
            for (int i = 0, vehCount = vehs.Count; i < vehCount; i++)
            {
                if (vehs[i].RegClass == regClass) { sales += vehs[i].Description.Sales[yrIndex]; }
            }
            return sales;
        }
        #endregion
        #region 
        public static void CalcFines(Manufacturer mfr, Parameters parameters)
        {
            Standards.CalcFines(mfr, parameters, RC.DomesticAuto);
            Standards.CalcFines(mfr, parameters, RC.ImportedAuto);
            Standards.CalcFines(mfr, parameters, RC.LightTruck  );
            Standards.CalcFines(mfr, parameters, RC.Unregulated );
        }
        public static void CalcFines(Manufacturer mfr, Parameters parameters, RC regClass)
        {
            Manufacturer.CModelData mmd = mfr.ModelData;
            double credits     = mmd.Credits    [regClass];     
            double tCreditsIn  = mmd.TCreditsIn [regClass];     
            double tCreditsOut = mmd.TCreditsOut[regClass];     
            double fineRate    = parameters.EconomicData.Kf;    
            mmd.Fines[regClass] = (double.IsNaN(credits)) ? 0 : -fineRate * Math.Min(Math.Round(credits + tCreditsIn - tCreditsOut, 1), 0);
        }
        public static RCDouble TestFines(Manufacturer mfr, Scenario scen, ModelYear year, Parameters parameters,
            RCDouble newCafe, TechImpact[] techImpacts, int techImpactCount)
        {
            RCDouble fines = new RCDouble();
            fines.DomesticAuto = Standards.TestFines(mfr, scen, year, parameters, RC.DomesticAuto, newCafe.DomesticAuto, techImpacts, techImpactCount);
            fines.ImportedAuto = Standards.TestFines(mfr, scen, year, parameters, RC.ImportedAuto, newCafe.ImportedAuto, techImpacts, techImpactCount);
            fines.LightTruck   = Standards.TestFines(mfr, scen, year, parameters, RC.LightTruck  , newCafe.LightTruck  , techImpacts, techImpactCount);
            fines.Unregulated  = Standards.TestFines(mfr, scen, year, parameters, RC.Unregulated , newCafe.Unregulated , techImpacts, techImpactCount);
            return fines;
        }
        public static double TestFines(Manufacturer mfr, Scenario scen, ModelYear year, Parameters parameters, RC regClass,
            double newCafe, TechImpact[] techImpacts, int techImpactCount)
        {
            double credits = Standards.TestCredits(mfr, scen, year, regClass, newCafe, techImpacts, techImpactCount);
            double fineRate = parameters.EconomicData.Kf;
            return -fineRate * Math.Min(credits, 0);
        }
        #endregion
        #region 
        public static void CalcCredits(Manufacturer mfr, Scenario scen, ModelYear year)
        {
            Standards.CalcCredits(mfr, scen, year, RC.DomesticAuto);
            Standards.CalcCredits(mfr, scen, year, RC.ImportedAuto);
            Standards.CalcCredits(mfr, scen, year, RC.LightTruck  );
            Standards.CalcCredits(mfr, scen, year, RC.Unregulated );
        }
        public static void CalcCredits(Manufacturer mfr, Scenario scen, ModelYear year, RC regClass)
        {
            Manufacturer.CModelData mmd = mfr.ModelData;
            double sales     = mmd.Sales   [regClass];
            double standard  = mmd.Standard[regClass];
            double cafe      = mmd.CAFE    [regClass];
            if (double.IsNaN(sales) || double.IsNaN(standard) || double.IsNaN(cafe)) { mmd.Credits[regClass] = 0; }
            else
            {   
                cafe += Standards.GetFFVCredits(mfr, scen, year, regClass);
                mmd.Credits[regClass] = Math.Round(sales * (Math.Round(cafe, 1) - standard), 1);
            }
        }
        public static RCDouble TestCredits(Manufacturer mfr, Scenario scen, ModelYear year, RCDouble newCafe,
            TechImpact[] techImpacts, int techImpactCount)
        {
            RCDouble credits = new RCDouble();
            credits.DomesticAuto = Standards.TestCredits(mfr, scen, year, RC.DomesticAuto, newCafe.DomesticAuto, techImpacts, techImpactCount);
            credits.ImportedAuto = Standards.TestCredits(mfr, scen, year, RC.ImportedAuto, newCafe.ImportedAuto, techImpacts, techImpactCount);
            credits.LightTruck   = Standards.TestCredits(mfr, scen, year, RC.LightTruck  , newCafe.LightTruck  , techImpacts, techImpactCount);
            credits.Unregulated  = Standards.TestCredits(mfr, scen, year, RC.Unregulated , newCafe.Unregulated , techImpacts, techImpactCount);
            return credits;
        }
        public static double TestCredits(Manufacturer mfr, Scenario scen, ModelYear year, RC regClass, double newCafe,
            TechImpact[] techImpacts, int techImpactCount)
        {
            Manufacturer.CModelData mmd = mfr.ModelData;
            double sales    = mmd.Sales[regClass];
            double standard = Standards.GetStandard(mfr, scen, year, regClass, techImpacts, techImpactCount);
            newCafe += Standards.GetFFVCredits(mfr, scen, year, regClass);
            return sales * (newCafe - standard);
        }
        static double GetFFVCredits(Manufacturer mfr, Scenario scen, ModelYear year, RC regClass)
        {
            if (mfr.Description.CreditsApplyToBaseline || !scen.IsBaseline)
            {
                return
                    (regClass == RC.DomesticAuto) ? mfr.Description.AvailableCreditsDA[year.Index] :
                    (regClass == RC.ImportedAuto) ? mfr.Description.AvailableCreditsIA[year.Index] :
                    (regClass == RC.LightTruck  ) ? mfr.Description.AvailableCreditsLT[year.Index] : 0;
            }
            return 0;
        }
        #endregion
        #region 
        public static void CalcCAFE(Manufacturer mfr, Scenario scen, ModelYear year, Parameters param)
        {
            Standards.CalcCAFE(mfr, scen, year, param, RegulatoryClass.DomesticAuto);
            Standards.CalcCAFE(mfr, scen, year, param, RegulatoryClass.ImportedAuto);
            Standards.CalcCAFE(mfr, scen, year, param, RegulatoryClass.LightTruck  );
            Standards.CalcCAFE(mfr, scen, year, param, RegulatoryClass.Unregulated );
        }
        public static void CalcCAFE(Manufacturer mfr, Scenario scen, ModelYear year, Parameters param, RC regClass)
        {
            double sales    = 0;
            double sofe     = 0;
            double sofe2Bag = 0;
            List<Vehicle> vehs = mfr.Vehicles;
            for (int i = 0, vehCount = vehs.Count; i < vehCount; i++)
            {
                Vehicle veh = vehs[i];
                if (veh.RegClass == regClass)
                {
                    Vehicle.CDescription vd = veh.Description;
                    double vehSales    = vd.Sales[year.Index];
                    double vehFE       = Math.Round(Standards.GetVehicleFuelEconomy(veh, scen, year, param, true , true ), 1);
                    double vehFE2Bag   = Math.Round(Standards.GetVehicleFuelEconomy(veh, scen, year, param, false, false), 1);
                    double vehSOFE     = vehSales / vehFE;
                    double vehSOFE2Bag = vehSales / vehFE2Bag;
                    sales    += vehSales;
                    sofe     += vehSOFE;
                    sofe2Bag += vehSOFE2Bag;
                    veh.ModelData.SalesOverFE = vehSOFE;
                } 
            } 
            mfr.ModelData.Sales      [regClass] = sales;
            mfr.ModelData.SalesOverFE[regClass] = sofe;
            mfr.ModelData.CAFE       [regClass] = sales / sofe;
            mfr.ModelData.CAFE_2Bag  [regClass] = sales / sofe2Bag;
        }
        public static RCDouble TestCAFE(Manufacturer mfr, Scenario scen, ModelYear year, Parameters param, TechImpact[] techImpacts,
            int techImpactCount, bool useRounding)
        {
            RCDouble cafe = new RCDouble();
            cafe.DomesticAuto = Standards.TestCAFE(mfr, scen, year, param, RC.DomesticAuto, techImpacts, techImpactCount, useRounding);
            cafe.ImportedAuto = Standards.TestCAFE(mfr, scen, year, param, RC.ImportedAuto, techImpacts, techImpactCount, useRounding);
            cafe.LightTruck   = Standards.TestCAFE(mfr, scen, year, param, RC.LightTruck  , techImpacts, techImpactCount, useRounding);
            cafe.Unregulated  = Standards.TestCAFE(mfr, scen, year, param, RC.Unregulated , techImpacts, techImpactCount, useRounding);
            return cafe;
        }
        public static double TestCAFE(Manufacturer mfr, Scenario scen, ModelYear year, Parameters param, RC regClass,
            TechImpact[] techImpacts, int techImpactCount, bool useRounding)
        {
            if (techImpactCount > 0 && (techImpacts == null || techImpacts.Length < techImpactCount))
            {   
                throw new ArgumentOutOfRangeException("techImpacts", "techImpacts is null or less than techImpactCount.");
            }
            Manufacturer.CModelData mmd = mfr.ModelData;
            double sales = mmd.Sales      [regClass];
            double sofe  = mmd.SalesOverFE[regClass];
            int multiFuelMode = scen.MultiFuelCAFE[year.Index];
            for (int i = 0; i < techImpactCount; i++)
            {
                if (techImpacts[i].Ignore) { continue; }
                Vehicle veh = techImpacts[i].Vehicle[year.Index];
                if (veh.RegClass == regClass)
                {
                    Vehicle.CDescription vd = veh.Description;
                    FuelValue adjFE    = Standards.GetAdjustedVehicleFuelEconomy(veh, techImpacts[i], scen, year, param, true, true);
                    double    vehFE    = Standards.GetVehicleFuelEconomy(adjFE, veh.Description.FuelShare, multiFuelMode);
                    double    vehSales = vd.Sales[year.Index];
                    double    vehSOFE  = veh.ModelData.SalesOverFE;
                    double    sofeDiff = vehSales / ((useRounding) ? Math.Round(vehFE, 1) : vehFE) - vehSOFE;
                    sofe += sofeDiff;
                }
            }
            return sales / sofe;
        }
        #endregion
        #region 
        public static double GetVehicleFuelEconomy(FuelValue fuelEconomy, FuelValue fuelShare, int mode)
        {
            if (fuelShare.IsMultiFuel)
            {   
                if (mode == 0 || (mode == 1 && fuelShare.Electricity == 0))
                {
                    return fuelEconomy.PrimaryValue;
                }
                else if ((mode == 1 || mode == 2) && fuelShare.Electricity > 0)
                {
                    return 1 / ((1 - fuelShare.Electricity) / fuelEconomy.PrimaryValue + fuelShare.Electricity / fuelEconomy.Electricity);
                }
                else if (mode == 2 && fuelShare.Ethanol85 > 0)
                {
                    return 1 / (fuelShare.Gasoline / fuelEconomy.Gasoline + fuelShare.Ethanol85 / fuelEconomy.Ethanol85);
                }
                throw new ArgumentException("The vehicle's fuel economy and/or fuel share values are not valid; or, incorrect multi-fuel mode specified in the scenario.");
            }
            else
            {   
                return fuelEconomy.PrimaryValue;
            }
        }
        public static double GetVehicleFuelEconomy(Vehicle veh, Scenario scen, ModelYear year, Parameters param,
            bool applyACAdjustment, bool applyOffCycleCredit)
        {
            FuelValue fuelEconomy = Standards.GetAdjustedVehicleFuelEconomy(veh, scen, year, param, applyACAdjustment, applyOffCycleCredit);
            return Standards.GetVehicleFuelEconomy(fuelEconomy, veh.Description.FuelShare, scen.MultiFuelCAFE[year.Index]);
        }
        public static FuelValue GetAdjustedVehicleFuelEconomy(Vehicle veh, Scenario scen, ModelYear year, Parameters param,
            bool applyACAdjustment, bool applyOffCycleCredit)
        {
            return Standards.GetAdjustedVehicleFuelEconomy(veh, null, scen, year, param, applyACAdjustment, applyOffCycleCredit);
        }
        static FuelValue GetAdjustedVehicleFuelEconomy(Vehicle veh, TechImpact techImpact,
            Scenario scen, ModelYear year, Parameters param, bool applyACAdjustment, bool applyOffCycleCredit)
        {
            double gramsCO2 = 0;
            if (applyACAdjustment && scen.IncludeAC[year.Index])
            {
                gramsCO2 += (veh.IsRCAuto) ? scen.PCACAdj[year.Index] : (veh.IsRCTruck) ? scen.LTACAdj[year.Index] : 0;
            }
            if (applyOffCycleCredit)
            {
                double cap = (veh.IsRCAuto) ? scen.PCOffCycleCap[year.Index] : (veh.IsRCTruck) ? scen.LTOffCycleCap[year.Index] : 0;
                double occ = veh.ModelData.OffCycleCredit;
                if (techImpact != null) { occ += techImpact.OffCycleCredit[year.Index]; }
                gramsCO2 += Math.Min(occ, cap);
            }
            FuelValue fuelEconomy = (techImpact == null) ? veh.Description.FuelEconomy : techImpact.NewFuelEconomy[year.Index];
            if (gramsCO2 == 0) { return fuelEconomy; }
            FuelValue fuelShare  = (techImpact == null) ? veh.Description.FuelShare : techImpact.NewFuelShare[year.Index];
            FuelValue fuelCarbon = Standards.GetFuelCarbonContent(param, true);
            if (fuelShare.Gasoline    > 0) { fuelEconomy[FuelType.Gasoline   ] = fuelCarbon.Gasoline    / (fuelCarbon.Gasoline    / fuelEconomy.Gasoline    - gramsCO2); }
            if (fuelShare.Diesel      > 0) { fuelEconomy[FuelType.Diesel     ] = fuelCarbon.Diesel      / (fuelCarbon.Diesel      / fuelEconomy.Diesel      - gramsCO2); }
            if (fuelShare.Ethanol85   > 0) { fuelEconomy[FuelType.Ethanol85  ] = fuelCarbon.Ethanol85   / (fuelCarbon.Ethanol85   / fuelEconomy.Ethanol85   - gramsCO2); }
            if (fuelShare.Electricity > 0) { fuelEconomy[FuelType.Electricity] = fuelCarbon.Electricity / (fuelCarbon.Electricity / fuelEconomy.Electricity - gramsCO2); }
            if (fuelShare.Hydrogen    > 0) { fuelEconomy[FuelType.Hydrogen   ] = fuelCarbon.Hydrogen    / (fuelCarbon.Hydrogen    / fuelEconomy.Hydrogen    - gramsCO2); }
            return fuelEconomy;
        }
        static FuelValue GetFuelCarbonContent(Parameters param, bool useGasolineForGEG)
        {
            FuelValue fuelCarbon = new FuelValue();
            FuelValue massDensity   = param.FuelProperties.MassDensity;
            FuelValue carbonContent = param.FuelProperties.CarbonContent;
            fuelCarbon[FuelType.Gasoline   ] = massDensity.Gasoline  * carbonContent.Gasoline  * 44 / 12;
            fuelCarbon[FuelType.Diesel     ] = massDensity.Diesel    * carbonContent.Diesel    * 44 / 12;
            fuelCarbon[FuelType.Ethanol85  ] = massDensity.Ethanol85 * carbonContent.Ethanol85 * 44 / 12;
            fuelCarbon[FuelType.Electricity] = (useGasolineForGEG) ? fuelCarbon.Gasoline : massDensity.Electricity * carbonContent.Electricity * 44 / 12;
            fuelCarbon[FuelType.Hydrogen   ] = (useGasolineForGEG) ? fuelCarbon.Gasoline : massDensity.Hydrogen    * carbonContent.Hydrogen    * 44 / 12;
            return fuelCarbon;
        }
        #endregion
        #region 
        public static void CalcVehicleTaxCredit(Industry data, Scenario scen, ModelYear year)
        {
            List<Manufacturer> mfrs = data.Manufacturers;
            for (int i = 0, mfrCount = mfrs.Count; i < mfrCount; i++)
            {
                List<Vehicle> vehs = mfrs[i].Vehicles;
                for (int j = 0, vehCount = vehs.Count; j < vehCount; j++)
                {
                    Vehicle veh = vehs[j];
                    veh.ModelData.TaxCredit = Standards.GetVehicleTaxCredit(veh, scen, year);
                }
            }
        }
        public static double GetVehicleTaxCredit(Vehicle veh, Scenario scen, ModelYear year)
        {
            return (veh.HEVType == HEVType.PureElectric) ? scen.TaxCreditEV  [year.Index] :
                   (veh.HEVType == HEVType.PlugInHybrid) ? scen.TaxCreditPHEV[year.Index] : 0;
        }
        #endregion
        #region 
        public static double GetVehicleFuelCost(FuelValue fuelEconomy, FuelValue fuelShare, FuelValue gap, KMYType kmy, int mode)
        {
            if (fuelShare.IsMultiFuel)
            {   
                if (mode == 0 || (mode == 1 && fuelShare.Electricity == 0))
                {
                    return kmy.Gasoline / (fuelEconomy.Gasoline * (1 - gap.Gasoline));
                }
                else if ((mode == 1 || mode == 2) && fuelShare.Electricity > 0)
                {
                    return kmy.Gasoline    / (fuelEconomy.Gasoline    * (1 - gap.Gasoline   )) * fuelShare.Gasoline +
                           kmy.Electricity / (fuelEconomy.Electricity * (1 - gap.Electricity)) * fuelShare.Electricity;
                }
                else if (mode == 2 && fuelShare.Ethanol85 > 0)
                {
                    return kmy.Gasoline  / (fuelEconomy.Gasoline  * (1 - gap.Gasoline )) * fuelShare.Gasoline +
                           kmy.Ethanol85 / (fuelEconomy.Ethanol85 * (1 - gap.Ethanol85)) * fuelShare.Ethanol85;
                }
                throw new ArgumentException("The vehicle's fuel economy and/or fuel share values are not valid; or, incorrect multi-fuel mode specified in the scenario.");
            }
            else
            {   
                return kmy[fuelEconomy.PrimaryFuel] / (fuelEconomy.PrimaryValue * (1 - gap[fuelEconomy.PrimaryFuel]));
            }
        }
        public static int GetCoefficients(RC regClass, Scenario scen, ModelYear year, out int fncType, out double[,] fncCoeff)
        {
            int yrIndex = year.Index;
            if (regClass == RC.DomesticAuto || regClass == RC.ImportedAuto)
            {
                fncType  = scen.AutoFunctionType[yrIndex];
                fncCoeff = scen.AutoCoefficients;
            }
            else if (regClass == RC.LightTruck)
            {
                fncType  = scen.LtFunctionType[yrIndex];
                fncCoeff = scen.LtCoefficients;
            }
            else
            {   
                fncType = 0;
                fncCoeff = null;
                return 0;
            }
            return fncCoeff.GetUpperBound(0) + 1;
        }
        #endregion
        #endregion
    }
}

